#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import ctypes

import numpy as np

from ciderpress.lib import load_library

libcider = load_library("libmcider")


class AtomicGridsIndexer:
    """
    This class is used to map between atomic grids ordering (which is used
    to integrate the theta functions for CIDER features) and sorted
    grids ordering (which is used to integrate the XC in PySCF).

    It also stores a variety of indexing arrays for performing integrals
    on individual atoms.

    Attributes:
        natm (int): number of atoms
        lmax (int): maximum l value for integration
        nlm (int): (lmax + 1)**2
        rad_arr (np.ndarray(np.float64), shape (nrad,)):
            Array of unique radial coordinates on the grid, sorted by atom.
        ar_loc (np.ndarray(np.int32), shape (nrad + 1,)):
            For each unique radial coordinate, this is the atom on which
            it is centered. I.e. the atom of rad_arr[r] is ar_loc[r]
        ra_loc (np.nddarray(np.int32), shape (natm + 1,)):
            For each atom a, ra_loc[a] is the first index in rad_arr that
            belongs to atom a, and ra_loc[a+1] - 1 is the last index that
            belongs to atom a.
        rad_loc (np.ndarray(np.int32), shape (nrad + 1,)):
            Specifies the range of indexes with a given radial coordinate.
            I.e. all_coords[rad_loc[r] : rad_loc[r+1]] have radial coordinate
            rad_arr[r].
        ylm (np.ndarray(np.float64), shape (num_ang_coord, nlm)):
            The stored spherical harmonics centered on each atom. Shape is
            the total number of unique angular coordinates X the number
            of spherical harmonics per coordinate (lmax + 1)**2.
        ylm_loc (np.ndarray(np.int32), shape (num_ang_coord,)):
            The spherical harmonics on all_coords[rad_loc[r] : rad_loc[r + 1]]
            are ylm[ylm_loc[r]]
        dirs (np.ndarray(np.float64), shape (num_ang_coord, 3)):
            Same as ylm, but just contains (x, y, z) instead of the spherical
            harmonics. Can be indexed by ylm_loc just like ylm
        all_weights (np.ndarray(np.float64)), shape (rad_loc[-1],)):
            All weights for the original, unsorted, unpruned grids, ordered
            by atom, then radial coordinate, then angular coordinate.
            Single flattened array for convenience.
        idx_map (np.ndarray(np.int32), shape (grids.weights.size,)):
            Indexing for map from a sorted/pruned grid to the unsorted,
            unpruned, atom-ordered grid.
            I.e. weights = all_weights[idx_map]
                 coords = all_coords[idx_map]
    """

    def __init__(self, natm, lmax, rad_arr, ar_loc, ra_loc, rad_loc, ylm, ylm_loc):
        """
        Initialize AtomicGridsIndexer. See attributes for argument details.

        Args:
            natm (int): Number of atoms
            lmax (int): Maximum l value
            rad_arr (np.ndarray): Radial coordinates
            ar_loc (np.ndarray): First radial coordinate index for each atom
            ra_loc (np.ndarray): Atom id of each radial coordinate
            rad_loc (np.ndarray): Location of each radial coordinate on
                the full grids.
            ylm (np.ndarray): Spherical harmonics
            ylm_loc (np.ndarray): Location of spherical harmonics for each
                radial coordinate.
        """
        self.natm = natm
        self.lmax = lmax
        self.nlm = (lmax + 1) * (lmax + 1)
        assert self.nlm == ylm.shape[1]
        self.rad_arr = rad_arr
        self.ar_loc = ar_loc
        self.ra_loc = ra_loc
        self.rad_loc = rad_loc
        self.ylm = ylm
        self.dirs = np.ascontiguousarray(ylm[:, [3, 1, 2]] * np.sqrt(4 * np.pi / 3))
        self.ylm_loc = ylm_loc
        self.all_weights = None
        self.idx_map = None
        self.iatom_list = None
        self.ga_loc = self.rad_loc[self.ra_loc]
        self.padding = 0

    @property
    def nrad(self):
        return self.rad_arr.size

    @property
    def ngrids(self):
        if self.all_weights is None:
            raise RuntimeError("weights must be set to get ngrids")
        return self.all_weights.size

    @classmethod
    def from_tabs(cls, mol, lmax, rad_loc_tab, ylm_loc_tab, rad_tab, ylm_tab):
        """
        Initialize AtomicGridsIndexer from tabulated elemental loc arrays.
        See Attributes for more details on what each component means.

        Args:
            mol (pyscf.gto.Mole): Molecule object
            lmax (int): Maximum l value
            rad_loc_tab (dict): Dictionary containing rad_loc array
                for each element in mol.
            ylm_loc_tab (dict): Dictionary containing ylm_loc array
                for each element in mol.
            rad_tab (dict): Dictionary containing array of radial
                coordinates for each element.
            ylm_tab (dict): Dictionary containing array of spherical
                harmonics for each element.

        Returns:
            AtomicGridsIndexer object
        """
        nlm = (lmax + 1) * (lmax + 1)
        full_rad_loc = np.array([0], dtype=np.int32)
        ar_loc = []
        ra_loc = [0]
        rads = []
        full_ylm_loc = np.array([], dtype=np.int32)
        full_ylm = np.empty((0, nlm), dtype=np.float64)
        for ia in range(mol.natm):
            symb = mol.atom_symbol(ia)
            nrad = rad_loc_tab[symb].size - 1
            start = full_rad_loc[-1]
            full_rad_loc = np.append(full_rad_loc, rad_loc_tab[symb][1:] + start)
            full_ylm_loc = np.append(
                full_ylm_loc, ylm_loc_tab[symb] + full_ylm.shape[0]
            )
            full_ylm = np.append(full_ylm, ylm_tab[symb], axis=0)
            rads.append(rad_tab[symb])
            rad_loc = full_rad_loc.size - 1
            ra_loc.append(rad_loc)
            ar_loc.append(ia * np.ones(nrad, dtype=np.int32))
        return cls(
            mol.natm,
            lmax,
            rad_arr=np.ascontiguousarray(np.concatenate(rads).astype(np.float64)),
            ar_loc=np.ascontiguousarray(np.concatenate(ar_loc).astype(np.int32)),
            ra_loc=np.array(ra_loc, dtype=np.int32, order="C"),
            rad_loc=np.ascontiguousarray(full_rad_loc.astype(np.int32)),
            ylm=np.ascontiguousarray(full_ylm.astype(np.float64)),
            ylm_loc=np.ascontiguousarray(full_ylm_loc.astype(np.int32)),
        )

    def set_padding(self, padding):
        """Padding is how much bigger target grid is than the grid
        constructed from all_coords[idx]"""
        self.padding = padding

    def set_weights(self, weights):
        """
        Set the atom-ordered, unsorted, unscreened weights.
        """
        self.all_weights = np.asarray(weights, order="C")

    def set_idx(self, idx):
        """
        Set the idx map between the atom-ordered grids and sorted grids.
        atom_ordered_grids[get_idx()] = sorted_screened_grids
        """
        self.idx_map = np.asarray(idx, order="C")
        tmp = np.arange(self.all_weights.size)
        for a in range(self.natm):
            tmp[self.ga_loc[a] : self.ga_loc[a + 1]] = a
        self.iatom_list = np.asarray(tmp[self.idx_map], order="C", dtype=np.int32)

    def get_idx(self):
        """
        Get the idx map between the atom-ordered grids and sorted grids.
        atom_ordered_grids[get_idx()] = sorted_screened_grids
        """
        return self.idx_map

    def empty_rlmq(self, nalpha=1, nspin=None):
        """
        Get an empty array for storying theta_rlmq.

        Args:
            nalpha (int): Number of interpolation points for CIDER kernel.

        Returns:
            np.ndarray with shape (nrad, nlm, nalpha)
        """
        if nspin is None:
            shape = (self.nrad, self.nlm, nalpha)
        else:
            shape = (nspin, self.nrad, self.nlm, nalpha)
        return np.empty(shape, order="C", dtype=np.float64)

    def empty_gq(self, nalpha=1, nspin=None):
        """
        Get an empty array for storing theta_gq.

        Args:
            nalpha (int): Number of interpolation points for CIDER kernel.

        Returns:
            np.ndarray with shape (ngrids, nalpha)
        """
        if nspin is None:
            shape = (self.ngrids, nalpha)
        else:
            shape = (nspin, self.ngrids, nalpha)
        return np.empty(shape, dtype=np.float64)

    def reduce_angc_ylm_(self, theta_rlmq, theta_gq, a2y=True, offset=None):
        """
        Project a function between angular coordinates and spherical harmonics.
        NOTE that this is an in-place operation (with the output being
        theta_rlmq if a2y=True and theta_gq if a2y=False), but the data
        in the output array does not need to be initialized, as it is
        overwritten on output, not added to.

        Args:
            theta_rlmq (np.ndarray): Shape (len(rad_arr), nlm, nalpha)
                Theta projected onto radial coordinates and spherical harmonics.
            theta_gq (np.ndarray): Shape (ngrids, nalpha)
            a2y (bool): If True, project real-space to spherical harmonics.
                If False, project spherical harmonics onto real-space coords.
        """
        if offset is None:
            offset = 0
        nalpha = theta_rlmq.shape[-1]
        stride = theta_gq.shape[-1]
        assert theta_rlmq.flags.c_contiguous
        assert theta_rlmq.shape == (self.nrad, self.nlm, nalpha)
        assert theta_gq.flags.c_contiguous
        assert theta_gq.shape == (self.all_weights.size, stride)
        assert theta_rlmq.dtype == np.float64
        assert theta_gq.dtype == np.float64
        assert nalpha + offset <= stride
        if a2y:
            fn = libcider.reduce_angc_to_ylm
        else:
            fn = libcider.reduce_ylm_to_angc
        fn(
            theta_rlmq.ctypes.data_as(ctypes.c_void_p),
            self.ylm.ctypes.data_as(ctypes.c_void_p),
            theta_gq.ctypes.data_as(ctypes.c_void_p),
            self.rad_loc.ctypes.data_as(ctypes.c_void_p),
            self.ylm_loc.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(nalpha),
            ctypes.c_int(self.nrad),
            ctypes.c_int(theta_gq.shape[0]),
            ctypes.c_int(self.nlm),
            ctypes.c_int(stride),
            ctypes.c_int(offset),
        )

    @classmethod
    def make_single_atom_indexer(cls, Y_nL, r_g):
        nn, nlm = Y_nL.shape
        lmax = int(np.sqrt(nlm + 1e-8)) - 1
        ng = r_g.size
        rad_arr = np.ascontiguousarray(r_g.astype(np.float64))
        ar_loc = np.zeros(ng, dtype=np.int32, order="C")
        ra_loc = np.asarray([0, ng], dtype=np.int32, order="C")
        rad_loc = np.ascontiguousarray((nn * np.arange(0, ng + 1)).astype(np.int32))
        ylm = np.ascontiguousarray(Y_nL.astype(np.float64))
        ylm_loc = np.zeros((ng,), dtype=np.int32)
        return cls(1, lmax, rad_arr, ar_loc, ra_loc, rad_loc, ylm, ylm_loc)



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

from ciderpress.dft.xc_evaluator import SplineSetEvaluator
from ciderpress.models.kernel_plans.kernel_tools import get_agpr_kernel
from ciderpress.models.kernel_plans.map_tools import get_mapped_gp_evaluator_additive


def get_kernel(
    natural_scale=None,
    natural_lscale=None,
    scale_factor=None,
    lscale_factor=None,
    **kwargs,
):
    # leave out density from feature vector
    slice(1, None, None)
    print(lscale_factor, natural_lscale)
    print(scale_factor, natural_scale)
    return get_agpr_kernel(
        slice(0, 1, None),
        slice(1, None, None),
        lscale_factor * natural_lscale,
        scale=[1e-5, 1e-5, scale_factor * natural_scale],
        order=2,
        nsingle=1,
    )


def mapping_plan(dft_kernel):
    scale, ind_sets, spline_grids, coeff_sets = get_mapped_gp_evaluator_additive(
        dft_kernel.kernel,
        dft_kernel.X1ctrl,
        dft_kernel.alpha,
        dft_kernel.feature_list,
    )
    # TODO this needs to get converted into a class
    # that evaluates a specific kernel's contribution to the
    # XC energy. This function should then be a component
    # called by DFTKernel.map().
    # Then these mapping plans need to be combined within some
    # overarching evaluator class, which should be created
    # by a function like MOLGP.map().
    return SplineSetEvaluator(scale, ind_sets, spline_grids, coeff_sets)



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import ctypes
from abc import ABC, abstractmethod

import numpy as np
import scipy.linalg.lapack
from pyscf import lib as pyscflib
from scipy.linalg import cho_factor, cho_solve, cholesky, solve_triangular
from scipy.special import gamma

from ciderpress import lib
from ciderpress.dft.settings import (
    NLDFSettings,
    SDMXBaseSettings,
    SDMXFullSettings,
    dalpha,
    ds2,
    get_alpha,
    get_cider_exponent,
    get_cider_exponent_gga,
    get_s2,
)
from ciderpress.lib import load_library as load_cider_library

libcider = load_cider_library("libmcider")


# map for C code
VJ_ID_MAP = {
    "se": 0,
    "se_ar2": 1,
    "se_a2r4": 2,
    "se_erf_rinv": 3,
}
VI_ID_MAP = {
    "se": 0,
    "se_r2": 1,
    "se_apr2": 2,
    "se_ap": 3,
    "se_ap2r2": 4,
    "se_lapl": 5,
    "se_rvec": 6,
    "se_grad": 7,
}


def _get_ovlp_fit_interpolation_coefficients(
    self, arg_g, i=-1, local=True, vbuf=None, dbuf=None
):
    """

    Args:
        self (NLDFAuxiliaryPlan):
        arg_g:
        i:
        local:
        vbuf:
        dbuf:

    Returns:

    """
    ngrids = arg_g.size
    p = self.empty_coefs(ngrids, local=local, buf=vbuf)
    dp = self.empty_coefs(ngrids, local=local, buf=dbuf)
    if i == -1:
        feat_id = VJ_ID_MAP["se"]
    else:
        assert 0 <= i < self.nldf_settings.num_feat_param_sets
        feat_id = VJ_ID_MAP[self.nldf_settings.feat_spec_list[i]]
    if feat_id == VJ_ID_MAP["se_erf_rinv"]:
        extra_args = self._get_extra_args(i)
        num_extra_args = len(extra_args)
        extra_args = (ctypes.c_double * num_extra_args)(*extra_args)
    else:
        extra_args = lib.c_null_ptr()
    alphas = self.local_alphas if local else self.alphas
    if self.coef_order == "gq":
        fn = libcider.cider_coefs_gto_gq
        shape = (arg_g.size, len(alphas))
        assert p.shape == shape
        assert dp.shape == shape
    else:
        fn = libcider.cider_coefs_gto_qg
        shape = (len(alphas), arg_g.size)
        assert p.shape == shape
        assert dp.shape == shape
    assert p.flags.c_contiguous
    assert dp.flags.c_contiguous
    assert arg_g.flags.c_contiguous
    assert alphas.flags.c_contiguous
    fn(
        p.ctypes.data_as(ctypes.c_void_p),
        dp.ctypes.data_as(ctypes.c_void_p),
        arg_g.ctypes.data_as(ctypes.c_void_p),
        alphas.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(arg_g.size),
        ctypes.c_int(len(alphas)),
        ctypes.c_int(feat_id),
        extra_args,
    )
    return p, dp


def _construct_cubic_splines(coefs_qi, dense_indexes):
    """
    Construct interpolating splines for NLDFAuxPlanSpline.

    The recipe is from

      https://en.wikipedia.org/wiki/Spline_(mathematics)

    Adapted from GPAW vdW implementation
    """
    q = dense_indexes
    n, N = coefs_qi.shape

    # shape N,n
    y = coefs_qi.T
    a = y
    h = q[1:] - q[:-1]
    # h = np.ones(q.size - 1) * (n - 1) / (N - 1)
    alpha = 3 * (
        (a[2:] - a[1:-1]) / h[1:, np.newaxis] - (a[1:-1] - a[:-2]) / h[:-1, np.newaxis]
    )
    l = np.ones((N, n))
    mu = np.zeros((N, n))
    z = np.zeros((N, n))
    for i in range(1, N - 1):
        l[i] = 2 * (q[i + 1] - q[i - 1]) - h[i - 1] * mu[i - 1]
        mu[i] = h[i] / l[i]
        z[i] = (alpha[i - 1] - h[i - 1] * z[i - 1]) / l[i]
    b = np.zeros((N, n))
    c = np.zeros((N, n))
    d = np.zeros((N, n))
    for i in range(N - 2, -1, -1):
        c[i] = z[i] - mu[i] * c[i + 1]
        b[i] = (a[i + 1] - a[i]) / h[i] - h[i] * (c[i + 1] + 2 * c[i]) / 3
        d[i] = (c[i + 1] - c[i]) / 3 / h[i]
    C_aip = np.zeros((n, N, 4))
    C_aip[:, :-1, 0] = a[:-1].T
    C_aip[:, :-1, 1] = b[:-1].T
    C_aip[:, :-1, 2] = c[:-1].T
    C_aip[:, :-1, 3] = d[:-1].T
    C_aip[-1, -1, 0] = 1.0
    return np.ascontiguousarray(C_aip.transpose(1, 0, 2))


def get_rho_tuple_with_grad_cross(rho_data, is_mgga=False):
    assert rho_data.ndim == 3
    drho = rho_data[:, 1:4]
    nspin = rho_data.shape[0]
    sigma = np.empty((2 * nspin - 1, rho_data.shape[-1]), order="F")
    sigma[::2] = np.einsum("sxg,sxg->sg", drho, drho)
    if nspin == 2:
        sigma[1] = np.einsum("xg,xg->g", drho[0], drho[1])
    rho = rho_data[:, 0].copy(order="F")
    if is_mgga:
        return rho, sigma, rho_data[:, 4].copy(order="F")
    else:
        return rho, sigma


def vxc_tuple_to_array(rho_data, vtuple):
    varr = np.zeros_like(rho_data)
    varr[:, 0] = vtuple[0]
    if len(vtuple) > 1:
        varr[:, 1:4] = 2 * vtuple[1][::2, None] * rho_data[:, 1:4]
        if len(vtuple) == 3:
            varr[:, 4] = vtuple[2]
        if rho_data.shape[0] == 2:
            varr[:, 1:4] += vtuple[1][1] * rho_data[::-1, 1:4]
    return varr


def get_rho_tuple(rho_data, with_spin=False, is_mgga=False):
    if with_spin:
        drho = rho_data[:, 1:4]
        sigma = np.einsum("sx...,sx...->s...", drho, drho)
        if is_mgga:
            return rho_data[:, 0].copy(), sigma, rho_data[:, 4].copy()
        else:
            return rho_data[:, 0].copy(), sigma
    else:
        drho = rho_data[1:4]
        sigma = np.einsum("x...,x...->...", drho, drho)
        if is_mgga:
            return rho_data[0], sigma, rho_data[4]
        else:
            return rho_data[0], sigma


def get_drhodf_tuple(rho_data, drhodf_data, with_spin=False, is_mgga=False):
    if with_spin:
        raise NotImplementedError("Have not implemented this function for with_spin")
        drho = rho_data[:, 1:4]
        ddrhodf = drhodf_data[:, 1:4]
        dsigmadf = 2 * np.einsum("sx...,sx...->s...", drho, ddrhodf)
        if is_mgga:
            return drhodf_data[:, 0].copy(), dsigmadf, drhodf_data[:, 4].copy()
        else:
            return drhodf_data[:, 0].copy(), dsigmadf
    else:
        drho = rho_data[1:4]
        ddrhodf = drhodf_data[1:4]
        dsigmadf = 2 * np.einsum("x...,x...->...", drho, ddrhodf)
        if is_mgga:
            return drhodf_data[0], dsigmadf, drhodf_data[4]
        else:
            return drhodf_data[0], dsigmadf


def get_ccl_settings(plan):
    has_vj = "j" in plan.nldf_settings.nldf_type
    ifeat_ids = []
    for spec in plan.nldf_settings.l0_feat_specs:
        ifeat_ids.append(VI_ID_MAP[spec])
    for spec in plan.nldf_settings.l1_feat_specs:
        ifeat_ids.append(VI_ID_MAP[spec])
    return has_vj, ifeat_ids


class _BaseSemilocalPlan:
    def __init__(self, settings, nspin):
        """
        Args:
            settings (SemilocalSettings): Settings for SL features
            nspin (int): Number of spin channels.
                1 for non-spin-polarized and 2 for spin-polarized
        """
        self.settings = settings
        self.nspin = nspin
        if self.settings.mode not in ["npa", "nst", "np", "ns"]:
            raise NotImplementedError

    @property
    def level(self):
        return self.settings.level

    def _fill_feat_npa_(self, rho, sigma, tau, feat):
        assert feat.shape[1] == 3
        feat[:, 0] = rho
        feat[:, 1] = get_s2(rho, sigma)
        feat[:, 2] = get_alpha(rho, sigma, tau)
        feat[:, 0] *= self.nspin
        feat[:, 1:3] /= self.nspin ** (2.0 / 3)

    def _fill_occd_npa_(
        self, feat, occd, rho, drho, sigma, dsigma, tau=None, dtau=None
    ):
        dpdn, dpdsigma = ds2(rho, sigma)
        feat[:, 0] = rho
        feat[:, 1] = get_s2(rho, sigma)
        occd[:, 0] = drho
        occd[:, 1] = dpdn * drho + dpdsigma * dsigma
        if self.level == "MGGA":
            feat[:, 2] = get_alpha(rho, sigma, tau)
            dadn, dadsigma, dadtau = dalpha(rho, sigma, tau)
            occd[:, 2] = dadn * drho + dadsigma * dsigma + dadtau * dtau
        feat[:, 0] *= self.nspin
        feat[:, 1:3] /= self.nspin ** (2.0 / 3)
        occd[:, 0] *= self.nspin
        occd[:, 1:3] /= self.nspin ** (2.0 / 3)

    def _fill_feat_nst_(self, rho, sigma, tau, feat):
        assert feat.shape[1] == 3
        feat[:, 0] = rho
        feat[:, 1] = sigma
        feat[:, 2] = tau
        feat[:, 0] *= self.nspin
        feat[:, 1] *= self.nspin * self.nspin
        feat[:, 2] *= self.nspin

    def _fill_occd_nst_(
        self, feat, occd, rho, drho, sigma, dsigma, tau=None, dtau=None
    ):
        feat[:, 0] = self.nspin * rho
        occd[:, 0] = self.nspin * drho
        feat[:, 1] = sigma * (self.nspin * self.nspin)
        occd[:, 1] = dsigma * (self.nspin * self.nspin)
        if self.level == "MGGA":
            feat[:, 2] = tau * self.nspin
            occd[:, 2] = dtau * self.nspin

    def _fill_feat_np_(self, rho, sigma, feat):
        assert feat.shape[1] == 2
        feat[:, 0] = rho
        feat[:, 1] = get_s2(rho, sigma)
        feat[:, 0] *= self.nspin
        feat[:, 1] /= self.nspin ** (2.0 / 3)

    def _fill_feat_ns_(self, rho, sigma, feat):
        assert feat.shape[1] == 2
        feat[:, 0] = rho
        feat[:, 1] = sigma
        feat[:, 0] *= self.nspin
        feat[:, 1] *= self.nspin * self.nspin

    def get_feat(self, rho, sigma, tau=None):
        for arr in [rho, sigma, tau]:
            assert arr is None or arr.ndim == 2
        feat = np.empty((rho.shape[0], self.settings.nfeat, rho.shape[1]))
        if self.settings.mode == "npa":
            self._fill_feat_npa_(rho, sigma, tau, feat)
        elif self.settings.mode == "nst":
            self._fill_feat_nst_(rho, sigma, tau, feat)
        elif self.settings.mode == "np":
            self._fill_feat_np_(rho, sigma, feat)
        else:
            self._fill_feat_ns_(rho, sigma, feat)
        return feat

    def get_occd(self, rho, drho, sigma, dsigma, tau=None, dtau=None):
        for arr in [rho, drho, sigma, dsigma, tau, dtau]:
            assert arr is None or arr.ndim == 2
        feat = np.empty((rho.shape[0], self.settings.nfeat, rho.shape[1]))
        occd = np.empty((rho.shape[0], self.settings.nfeat, rho.shape[1]))
        if self.settings.mode in ["npa", "np"]:
            self._fill_occd_npa_(feat, occd, rho, drho, sigma, dsigma, tau, dtau)
        elif self.settings.mode in ["nst", "ns"]:
            self._fill_occd_nst_(feat, occd, rho, drho, sigma, dsigma, tau, dtau)
        return feat, occd


class SemilocalPlan(_BaseSemilocalPlan):
    def get_feat(self, rho):
        sigma = np.einsum("sx...,sx...->s...", rho[:, 1:4], rho[:, 1:4])
        if self.level == "MGGA":
            tau = rho[:, 4]
        else:
            tau = None
        return super(SemilocalPlan, self).get_feat(rho[:, 0], sigma, tau)

    def get_occd(self, rho, drho):
        assert rho.ndim == drho.ndim == 3
        sigma = np.einsum("sx...,sx...->s...", rho[:, 1:4], rho[:, 1:4])
        dsigma = 2 * np.einsum("sx...,sx...->s...", rho[:, 1:4], drho[:, 1:4])
        if self.level == "MGGA":
            tau = rho[:, 4]
            dtau = drho[:, 4]
        else:
            tau = None
            dtau = None
        return super(SemilocalPlan, self).get_occd(
            rho[:, 0], drho[:, 0], sigma, dsigma, tau, dtau
        )

    def _fill_vxc_npa_(self, rho, vfeat, vxc):
        sigma = np.einsum("sx...,sx...->s...", rho[:, 1:4], rho[:, 1:4])
        sm23 = self.nspin ** (-2.0 / 3)
        dpdn, dpdsigma = ds2(rho[:, 0], sigma)
        dadn, dadsigma, dadtau = dalpha(rho[:, 0], sigma, rho[:, 4])
        vxc[:, 0] += self.nspin * vfeat[:, 0] + sm23 * (
            vfeat[:, 1] * dpdn + vfeat[:, 2] * dadn
        )
        vxc[:, 1:4] += (
            (sm23 * (vfeat[:, 1] * dpdsigma + vfeat[:, 2] * dadsigma))[:, None, :]
            * 2
            * rho[:, 1:4]
        )
        vxc[:, 4] += sm23 * vfeat[:, 2] * dadtau

    def _fill_vxc_nst_(self, rho, vfeat, vxc):
        vxc[:, 0] += self.nspin * vfeat[:, 0]
        # fmt: off
        vxc[:, 1:4] += (
            (self.nspin * self.nspin * 2 * vfeat[:, 1])[:, None, :]
            * rho[:, 1:4]
        )
        # fmt: on
        vxc[:, 4] += self.nspin * vfeat[:, 2]

    def _fill_vxc_np_(self, rho, vfeat, vxc):
        sigma = np.einsum("sx...,sx...->s...", rho[:, 1:4], rho[:, 1:4])
        sm23 = self.nspin ** (-2.0 / 3)
        dpdn, dpdsigma = ds2(rho[:, 0], sigma)
        vxc[:, 0] += self.nspin * vfeat[:, 0] + sm23 * vfeat[:, 1] * dpdn
        vxc[:, 1:4] += (sm23 * vfeat[:, 1] * dpdsigma)[:, None, :] * 2 * rho[:, 1:4]

    def _fill_vxc_ns_(self, rho, vfeat, vxc):
        vxc[:, 0] += self.nspin * vfeat[:, 0]
        # fmt: off
        vxc[:, 1:4] += (
            (self.nspin * self.nspin * 2 * vfeat[:, 1])[:, None, :]
            * rho[:, 1:4]
        )
        # fmt: on

    def get_vxc(self, rho, vfeat, vxc=None):
        assert rho.ndim == 3
        if vxc is None:
            vxc = np.zeros_like(rho[:, :5])
        if self.settings.mode == "npa":
            self._fill_vxc_npa_(rho, vfeat, vxc)
        elif self.settings.mode == "nst":
            self._fill_vxc_nst_(rho, vfeat, vxc)
        elif self.settings.mode == "np":
            self._fill_vxc_np_(rho, vfeat, vxc)
        else:
            self._fill_vxc_ns_(rho, vfeat, vxc)
        return vxc


class SemilocalPlan2(_BaseSemilocalPlan):
    def _fill_vxc_npa_(self, vfeat, rho, vrho, sigma, vsigma, tau=None, vtau=None):
        sm23 = self.nspin ** (-2.0 / 3)
        dpdn, dpdsigma = ds2(rho, sigma)
        if self.level == "MGGA":
            dadn, dadsigma, dadtau = dalpha(rho, sigma, tau)
            vrho[:] += self.nspin * vfeat[:, 0] + sm23 * (
                vfeat[:, 1] * dpdn + vfeat[:, 2] * dadn
            )
            vsigma[:] += sm23 * (vfeat[:, 1] * dpdsigma + vfeat[:, 2] * dadsigma)
            vtau[:] += sm23 * vfeat[:, 2] * dadtau
        else:
            vrho[:] += self.nspin * vfeat[:, 0] + sm23 * vfeat[:, 1] * dpdn
            vsigma[:] += sm23 * vfeat[:, 1] * dpdsigma

    def _fill_vxc_nst_(self, vfeat, vrho, vsigma, vtau=None):
        vrho[:] += self.nspin * vfeat[:, 0]
        vsigma[:] += self.nspin * self.nspin * vfeat[:, 1]
        if self.level == "MGGA":
            vtau[:] += self.nspin * vfeat[:, 2]

    def get_vxc(self, vfeat, rho, vrho, sigma, vsigma, tau=None, vtau=None):
        for arr in [rho, vrho, sigma, vsigma, tau, vtau]:
            assert arr is None or arr.ndim == 2
        if self.settings.mode in ["npa", "np"]:
            self._fill_vxc_npa_(vfeat, rho, vrho, sigma, vsigma, tau, vtau)
        elif self.settings.mode in ["nst", "ns"]:
            self._fill_vxc_nst_(vfeat, vrho, vsigma, vtau)


class FracLaplPlan:
    """
    Plan for the Fractional Laplacian features. These features do not
    require additional settings for how to compute them, so this
    class basically just stores the settings class and nspin. Currently,
    evaluation of the orbital operations must be implemented
    in each DFT code separately.
    """

    def __init__(self, settings, nspin, nrho_sl=5, drho_start=1):
        """
        Initialize FracLaplPlan instance

        Args:
            settings (FracLaplSettings): settings for the fractional
                laplacian descriptors
        """
        self.settings = settings
        self.nspin = nspin
        self._cached_l1_data = []
        self._cached_ld_data = []
        self._nsl = nrho_sl
        self._i0d = drho_start

    def _clear_l1_cache(self):
        self._cached_l1_data = []
        self._cached_ld_data = []

    def _cache_l1_vectors(self, drho, l1_data, make_copy):
        assert len(l1_data[0]) % 3 == 0
        assert len(l1_data[0]) // 3 == self.settings.nk1
        for i in range(self.settings.nk1):
            self._cached_l1_data.append(l1_data[:, 3 * i : 3 * i + 3])
        self._cached_l1_data.append(drho)
        if make_copy:
            self._cached_l1_data = [v.copy() for v in self._cached_l1_data]

    def _cache_ld_vectors(self, drho, l1_data, make_copy):
        assert len(l1_data[0]) % 3 == 0
        assert len(l1_data[0]) // 3 == self.settings.nd1
        for i in range(self.settings.nk1):
            self._cached_ld_data.append(l1_data[:, 3 * i : 3 * i + 3])
        self._cached_ld_data.append(drho)
        if make_copy:
            self._cached_ld_data = [v.copy() for v in self._cached_ld_data]

    def _cache_all_l1_data(self, rho_data, make_l1_data_copy):
        nk0 = self.settings.nk0
        nk1 = self.settings.nk1
        nd1 = self.settings.nd1
        self._cache_l1_vectors(
            rho_data[:, self._i0d : self._i0d + 3],
            rho_data[:, self._nsl + nk0 : self._nsl + nk0 + 3 * nk1],
            make_l1_data_copy,
        )
        start = self._nsl + nk0 + 3 * nk1
        self._cache_ld_vectors(
            rho_data[:, self._i0d : self._i0d + 3],
            rho_data[:, start : start + 3 * nd1],
            make_l1_data_copy,
        )

    def get_feat(self, rho_data, feat=None, make_l1_data_copy=True):
        self._clear_l1_cache()
        if rho_data.ndim == 2:
            rho_data = rho_data[None, :, :]
        nk0 = self.settings.nk0
        nspin = self.nspin
        assert rho_data.ndim == 3
        assert rho_data.shape[:2] == (nspin, self._nsl + self.settings.nrho)
        l0_data = rho_data[:, self._nsl : self._nsl + nk0]
        self._cache_all_l1_data(rho_data, make_l1_data_copy)
        ngrids = rho_data.shape[2]
        if feat is None:
            feat = np.empty((nspin, self.settings.nfeat, ngrids))
        assert feat.shape == (nspin, self.settings.nfeat, ngrids)
        for i in range(nk0):
            feat[:, i] = l0_data[:, i]
        for i1 in range(len(self.settings.l1_dots)):
            j, k = self.settings.l1_dots[i1]
            feat[:, nk0 + i1] = np.einsum(
                "sxg,sxg->sg", self._cached_l1_data[j], self._cached_l1_data[k]
            )
        nl1 = len(self.settings.l1_dots)
        for i1 in range(len(self.settings.ld_dots)):
            j, k = self.settings.ld_dots[i1]
            feat[:, nk0 + nl1 + i1] = np.einsum(
                "sxg,sxg->sg", self._cached_ld_data[j], self._cached_ld_data[k]
            )
        ndd = self.settings.ndd
        for i in range(ndd):
            feat[:, i - ndd] = rho_data[:, i - ndd]
        return feat

    def get_occd(self, rho_data, occd_data):
        assert rho_data.ndim == 2
        assert self.nspin == 1
        assert occd_data.ndim == 3
        assert occd_data.shape[1:] == rho_data.shape
        norb = occd_data.shape[0]
        nk0 = self.settings.nk0
        ngrids = rho_data.shape[1]
        l0_occd = occd_data[:, self._nsl : self._nsl + nk0]
        self._clear_l1_cache()
        self._cache_all_l1_data(rho_data[None, :], False)
        l1_rho_cache = [v for v in self._cached_l1_data]
        ld_rho_cache = [v for v in self._cached_ld_data]
        self._clear_l1_cache()
        self._cache_all_l1_data(occd_data, False)
        l1_occd_cache = [v for v in self._cached_l1_data]
        ld_occd_cache = [v for v in self._cached_ld_data]
        self._clear_l1_cache()
        occd_feat = np.empty((norb, self.settings.nfeat, ngrids))
        for i in range(nk0):
            occd_feat[:, i] = l0_occd[:, i]
        for i1 in range(len(self.settings.l1_dots)):
            j, k = self.settings.l1_dots[i1]
            occd_feat[:, nk0 + i1] = np.einsum(
                "sxg,oxg->og", l1_rho_cache[j], l1_occd_cache[k]
            ) + np.einsum("sxg,oxg->og", l1_rho_cache[k], l1_occd_cache[j])
        nl1 = len(self.settings.l1_dots)
        for i1 in range(len(self.settings.ld_dots)):
            j, k = self.settings.ld_dots[i1]
            occd_feat[:, nk0 + nl1 + i1] = np.einsum(
                "sxg,oxg->og", ld_rho_cache[j], ld_occd_cache[k]
            ) + np.einsum("sxg,oxg->og", ld_rho_cache[k], ld_occd_cache[j])
        ndd = self.settings.ndd
        for i in range(ndd):
            occd_feat[:, i - ndd] = occd_data[:, i - ndd]
        return occd_feat

    def get_vxc(self, vfeat, vxc=None):
        if vfeat.ndim == 2:
            vfeat = vfeat[None, :, :]
        nk0 = self.settings.nk0
        nk1 = self.settings.nk1
        nd1 = self.settings.nd1
        nspin = self.nspin
        assert vfeat.ndim == 3
        assert vfeat.shape[:2] == (nspin, self.settings.nfeat)
        ngrids = vfeat.shape[2]
        if vxc is None:
            vxc = np.zeros((nspin, self._nsl + self.settings.nrho, ngrids))
        assert vxc.shape == (nspin, self._nsl + self.settings.nrho, ngrids)
        vdrho = vxc[:, self._i0d : self._i0d + 3]
        l0_vxc = vxc[:, self._nsl : self._nsl + nk0]
        l1_vxc = vxc[:, self._nsl + nk0 : self._nsl + nk0 + 3 * nk1]
        start = self._nsl + nk0 + 3 * nk1
        ld_vxc = vxc[:, start : start + 3 * nd1]
        ldd_vxc = vxc[:, start + 3 * nd1 :]
        for i in range(nk0):
            l0_vxc[:, i] += vfeat[:, i]
        for i1 in range(len(self.settings.l1_dots)):
            j, k = self.settings.l1_dots[i1]
            vfeat_tmp = vfeat[:, nk0 + i1 : nk0 + i1 + 1, :]
            if j == -1:
                target = vdrho
            else:
                target = l1_vxc[:, 3 * j : 3 * j + 3]
            target[:] += vfeat_tmp * self._cached_l1_data[k]
            if k == -1:
                target = vdrho
            else:
                target = l1_vxc[:, 3 * k : 3 * k + 3]
            target[:] += vfeat_tmp * self._cached_l1_data[j]
        start = nk0 + len(self.settings.l1_dots)
        for i1 in range(len(self.settings.ld_dots)):
            j, k = self.settings.ld_dots[i1]
            vfeat_tmp = vfeat[:, start + i1 : start + i1 + 1, :]
            if j == -1:
                target = vdrho
            else:
                target = ld_vxc[:, 3 * j : 3 * j + 3]
            target[:] += vfeat_tmp * self._cached_ld_data[k]
            if k == -1:
                target = vdrho
            else:
                target = ld_vxc[:, 3 * k : 3 * k + 3]
            target[:] += vfeat_tmp * self._cached_ld_data[j]
        ndd = self.settings.ndd
        for i in range(ndd):
            ldd_vxc[:, i] += vfeat[:, i - ndd]
        return vxc


class SADMPlan:
    """
    Plan for evaluating exchange energy of spherically averaged density
    matrix. Stores the settings (basically an empty class since the feature
    is not tunable, though local range separation might be added later),
    the nspin, and paramters for an even-tempered Gaussian basis for
    expanding the spherically averaged density matrix. Currently,
    evaluation of the integrals must be implemented in each DFT code separately.
    """

    has_coul_list = False

    def __init__(self, settings, nspin, alpha0, lambd, nalpha, fit_metric="ovlp"):
        """

        Args:
            settings (SADMSettings): Settings for the descriptor.
            nspin (int): Number of spins, 2 for spin-polarized and 1 for
                non-spin-polarized
            alpha0 (float): Minimum exponent. sqrt(1/alpha0) should be at least
                the width of the system, or altneratively the maximum distance
                at which the exchange hole is expected to have significant density.
            lambd (float): ETB lambda parameter
            nalpha (int): Number of ETB basis functions
            fit_metric (str): 'ovlp' or 'coul', the metric for expanding the
                density matrix in the auxiliary basis.
        """
        self.settings = settings
        self.nspin = nspin
        if alpha0 <= 0:
            raise ValueError("alpha0 must be positive")
        self.alpha0 = alpha0
        if lambd <= 1:
            raise ValueError("lambd must be > 1")
        self.lambd = lambd
        if not isinstance(nalpha, int) or nalpha <= 0:
            raise ValueError("nalpha must be positive integer")
        self.nalpha = nalpha
        self.alphas = self.alpha0 * self.lambd ** np.arange(self.nalpha)
        self.alpha_norms = (np.pi / (2 * self.alphas)) ** -0.75
        prod = self.alphas * self.alphas[:, None]
        sum = self.alphas + self.alphas[:, None]
        # coul = ( alpha(r) | 1/(r-r') | beta(r') )
        # coul = 4 * np.sqrt(2) * np.pi / prod**0.25 / sum**0.5
        # coul = ( alpha(r) | 1 / r | beta(r) )
        coul = 4 * np.sqrt(2 / np.pi) * prod**0.75 / sum
        LJ = cholesky(coul, lower=True)
        if settings.mode == "smooth":
            fit_metric = "ovlp"
            r2vals = 2.0 / self.alphas
            vals = np.exp(-self.alphas[None, :] * r2vals[:, None])
            vals *= self.alpha_norms[None, :]
            self.fit_matrix = np.linalg.solve(vals.T, LJ).T
            self.alpha_norms = (
                4 * self.alphas**1.5 / ((4 - np.sqrt(2)) * np.pi**1.5)
            )
        elif fit_metric == "ovlp":
            # < alpha(r) | beta(r) >
            ovlp = 2 * np.sqrt(2) * prod**0.75 / sum**1.5
            c_and_l = cho_factor(ovlp)
            fit = cho_solve(c_and_l, LJ)
            self.fit_matrix = np.ascontiguousarray(fit.T)
        elif fit_metric == "coul":
            self.fit_matrix = solve_triangular(LJ, np.identity(LJ.shape[0]), lower=True)
        else:
            raise ValueError('fit_metric must be "ovlp" or "coul"')
        self.fit_metric = fit_metric

    @property
    def num_l0_feat(self):
        return 1

    @property
    def num_l1_feat(self):
        return 0

    def get_features(self, p_vag, out=None, l0tmp=None, l1tmp=None):
        """
        Should pass out[idm], l0tmp[idm], etc. to this function
        """
        na, ng = p_vag.shape[1:]
        nfeat = self.settings.nfeat
        n0 = self.num_l0_feat
        assert n0 == 1
        assert nfeat == 1
        if out is None:
            out = np.empty((nfeat, ng))
        if l0tmp is None:
            l0tmp = np.empty((n0, na, ng))
        l0tmp[0] = pyscflib.dot(self.fit_matrix, p_vag[0])
        out[0] = pyscflib.einsum("qg,qg->g", l0tmp[0], l0tmp[0])
        out[0] *= -0.25 * self.nspin * self.nspin
        return out

    def get_vxc(self, vxc_ig, l0tmp, l1tmp=None, out=None):
        nfeat = self.settings.nfeat
        n0 = self.num_l0_feat
        assert nfeat == n0 == vxc_ig.shape[0] == 1
        ng = vxc_ig.shape[1]
        if out is None:
            out = np.zeros((1, self.nalpha, ng))
        else:
            out[:] = 0.0
        if nfeat == 0:
            return out
        tmp = -0.25 * vxc_ig[:n0, None] * l0tmp * self.nspin * self.nspin  # fqg
        out[0] += pyscflib.dot(self.fit_matrix.T, tmp[0])
        return out


def _sdmx_einsum(tmp, mystr=None):
    if mystr is not None:
        pass
    elif tmp.ndim == 2:
        mystr = "qg,qg->g"
    elif tmp.ndim == 3:
        mystr = "vqg,vqg->g"
    else:
        raise ValueError
    if tmp.dtype == np.float64:
        return pyscflib.einsum(mystr, tmp, tmp)
    elif tmp.dtype == np.complex128:
        return pyscflib.einsum(mystr, tmp.real, tmp.real) + pyscflib.einsum(
            mystr, tmp.imag, tmp.imag
        )


class SDMXBasePlan:

    has_coul_list: bool = True
    nspin: int
    nalpha: int
    settings: SDMXBaseSettings
    fit_matrices: list
    alphas_norms = None
    alphas = None

    @property
    def num_l0_feat(self):
        raise NotImplementedError

    @property
    def num_l1_feat(self):
        raise NotImplementedError

    def _get_fit_mats(self):
        if self.has_coul_list:
            fit_mats = self.fit_matrices
        else:
            fit_mats = [self.fit_matrix]
        return fit_mats

    def get_features(self, p_vag, out=None, l0tmp=None, l1tmp=None):
        """
        Should pass out[idm], l0tmp[idm], etc. to this function
        """
        na, ng = p_vag.shape[1:]
        nfeat = self.settings.nfeat
        n0 = self.num_l0_feat
        if out is None:
            out = np.empty((nfeat, ng))
        if l0tmp is None:
            l0tmp = np.empty((n0, na, ng))
        elif l1tmp is None:
            l1tmp = np.empty((nfeat - n0, 3, na, ng))
        fit_mats = self._get_fit_mats()
        fac = -0.25 * self.nspin * self.nspin
        for ifeat, fit_matrix in enumerate(fit_mats[:n0]):
            l0tmp[ifeat] = pyscflib.dot(fit_matrix, p_vag[0])
            # out[ifeat] = pyscflib.einsum(
            #     'qg,qg->g', l0tmp[ifeat], l0tmp[ifeat]
            # )
            out[ifeat] = _sdmx_einsum(l0tmp[ifeat])
            out[ifeat] *= fac
        for ifeat, fit_matrix in enumerate(fit_mats[n0:]):
            ifeat_shift = ifeat + n0
            for v in range(3):
                l1tmp[ifeat, v] = pyscflib.dot(fit_matrix, p_vag[v + 1])
            # out[ifeat_shift] = pyscflib.einsum(
            #     'vqg,vqg->g', l1tmp[ifeat], l1tmp[ifeat]
            # )
            out[ifeat_shift] = _sdmx_einsum(l1tmp[ifeat])
            out[ifeat_shift] *= fac  # factor of -1/4 for EXX
        return out

    def get_vxc(self, vxc_ig, l0tmp, l1tmp=None, out=None):
        nfeat = self.settings.nfeat
        assert nfeat == vxc_ig.shape[0]
        ng = vxc_ig.shape[1]
        n0 = self.num_l0_feat
        n1 = self.num_l1_feat
        if n1 == 0:
            nv = 1
        else:
            nv = 4
            assert l1tmp is not None
        if out is None:
            out = np.zeros((nv, self.nalpha, ng), dtype=l0tmp.dtype)
        else:
            out[:] = 0.0
        if nfeat == 0:
            return out
        fit_mats = self._get_fit_mats()
        fac = -0.25 * self.nspin * self.nspin
        tmp = fac * vxc_ig[:n0, None] * l0tmp  # fqg
        # if tmp.dtype == np.complex128:
        #     tmp = tmp.conj()
        for ifeat, fit_matrix in enumerate(fit_mats[:n0]):
            out[0] += pyscflib.dot(fit_matrix.T, tmp[ifeat])
        if n1 > 0:
            tmp1 = fac * vxc_ig[n0:, None, None] * l1tmp
            # if tmp1.dtype == np.complex128:
            #     tmp1 = tmp1.conj()
            for ifeat, fit_matrix in enumerate(self._get_fit_mats()[n0:]):
                for v in range(3):
                    # tmp2[:, v + 1] += fit_matrix.T.dot(tmp1[ifeat, v])
                    out[v + 1] += pyscflib.dot(fit_matrix.T, tmp1[ifeat, v])
        return out


class SDMXPlan(SDMXBasePlan):

    has_coul_list = True

    def __init__(self, settings, nspin, alpha0, lambd, nalpha, fit_metric="ovlp"):
        """

        Args:
            settings (SDMXSettings): Settings for the descriptor.
            nspin (int): Number of spins, 2 for spin-polarized and 1 for
                non-spin-polarized
            alpha0 (float): Minimum exponent. sqrt(1/alpha0) should be at least
                the width of the system, or altneratively the maximum distance
                at which the exchange hole is expected to have significant density.
            lambd (float): ETB lambda parameter
            nalpha (int): Number of ETB basis functions
            fit_metric (str): 'ovlp' or 'coul', the metric for expanding the
                density matrix in the auxiliary basis.
        """
        self.settings = settings
        self.nspin = nspin
        if alpha0 <= 0:
            raise ValueError("alpha0 must be positive")
        self.alpha0 = alpha0
        if lambd <= 1:
            raise ValueError("lambd must be > 1")
        self.lambd = lambd
        if not isinstance(nalpha, int) or nalpha <= 0:
            raise ValueError("nalpha must be positive integer")
        self.nalpha = nalpha
        self.alphas = self.alpha0 * self.lambd ** np.arange(self.nalpha)
        self.alpha_norms = (np.pi / (2 * self.alphas)) ** -0.75
        self.alpha_norms_l1 = np.sqrt(4.0 / 3 * self.alphas) * self.alpha_norms
        prod = self.alphas * self.alphas[:, None]
        sum = self.alphas + self.alphas[:, None]
        # coul = ( alpha(r) | 1 / r | beta(r) )
        coul_list = [
            4
            * np.sqrt(2 / np.pi)
            * prod**0.75
            / sum ** (0.5 * (3 - n))
            * gamma(0.5 * (3 - n))
            for n in settings.pows
        ]
        ndt = self.settings.ndterms
        if ndt > 0:
            coul_list += [
                16
                * np.sqrt(2 / np.pi)
                * prod**1.75
                / sum ** (0.5 * (7 - n))
                * gamma(0.5 * (7 - n))
                for n in settings.pows[:ndt]
            ]
        n1t = self.settings.n1terms
        if n1t > 0:
            coul_list += [
                4
                * np.sqrt(2 / np.pi)
                * prod**0.75
                / sum ** (0.5 * (5 - n))
                * gamma(0.5 * (5 - n))
                for n in settings.pows[:n1t]
            ]
        LJ_list = [cholesky(coul, lower=True) for coul in coul_list]
        fit_metric = "ovlp"
        if settings.mode == "smooth":
            r2vals = 2.0 / self.alphas
            vals = np.exp(-self.alphas[None, :] * r2vals[:, None])
            vals *= self.alpha_norms[None, :]
            self.fit_matrices = [
                np.ascontiguousarray(np.linalg.solve(vals.T, LJ).T) for LJ in LJ_list
            ]
            self.alpha_norms = (
                4 * self.alphas**1.5 / ((4 - np.sqrt(2)) * np.pi**1.5)
            )
        else:
            # < alpha(r) | beta(r) >
            ovlp = 2 * np.sqrt(2) * prod**0.75 / sum**1.5
            c_and_l = cho_factor(ovlp)
            self.fit_matrices = [
                np.ascontiguousarray(cho_solve(c_and_l, LJ).T) for LJ in LJ_list
            ]
            ovlp = 4 * np.sqrt(2) * prod**1.25 / sum**2.5
        self.fit_metric = fit_metric

    @property
    def num_l0_feat(self):
        if self.settings.n1terms == 0:
            return self.settings.nfeat
        else:
            return len(self.settings.pows) + self.settings.ndterms

    @property
    def num_l1_feat(self):
        return self.settings.nfeat - self.num_l0_feat


def _get_int_0(n, prod, asum):
    return (
        4
        * np.sqrt(2 / np.pi)
        * prod**0.75
        / asum ** (0.5 * (3 - n))
        * gamma(0.5 * (3 - n))
    )


def _get_int_d(n, prod, asum):
    return (
        16
        * np.sqrt(2 / np.pi)
        * prod**1.75
        / asum ** (0.5 * (7 - n))
        * gamma(0.5 * (7 - n))
    )


def _get_int_1(n, prod, asum):
    return (
        4
        * np.sqrt(2 / np.pi)
        * prod**0.75
        / asum ** (0.5 * (5 - n))
        * gamma(0.5 * (5 - n))
    )


def _get_int_1d(n, prod, asum):
    return (
        4
        * np.sqrt(2 / np.pi)
        * prod**0.75
        / asum ** (0.5 * (9 - n))
        * gamma(0.5 * (5 - n))
        * ((n - 4) * asum**2 + prod * (35 - 12 * n + n * n))
    )


class SDMXFullPlan(SDMXBasePlan):

    settings: SDMXFullSettings
    fit_metric = "ovlp"

    def __init__(
        self,
        settings,
        nspin,
        alpha0,
        lambd,
        nalpha,
    ):
        """

        Args:
            settings (SDMXFullSettings):
            nspin (int):
            alpha0 (float):
            lambd (float):
            nalpha (int):
        """
        self.settings = settings
        self.nspin = nspin
        if alpha0 <= 0:
            raise ValueError("alpha0 must be positive")
        self.alpha0 = alpha0
        if lambd <= 1:
            raise ValueError("lambd must be > 1")
        self.lambd = lambd
        if not isinstance(nalpha, int) or nalpha <= 0:
            raise ValueError("nalpha must be positive integer")
        self.nalpha = nalpha
        self.alphas = self.alpha0 * self.lambd ** np.arange(self.nalpha)
        self.alpha_norms = (np.pi / (2 * self.alphas)) ** -0.75
        l0mats = []
        l1mats = []
        for ratio in self.settings.ratios:
            r = ratio
            prod = self.alphas * self.alphas[:, None]
            isum = self.alphas + self.alphas[:, None]
            asum = self.alphas / r + r * self.alphas[:, None]
            bsum = self.alphas * r + self.alphas[:, None] / r
            for n, rdr in self.settings.iterate_l0_terms(ratio):
                # num = 0.25 * (ratio ** (0.5 * n) + ratio ** (-0.5 * n))
                num = 0.5
                if not rdr:
                    l0mats.append(
                        num * _get_int_0(n, prod, isum)
                        + 0.25 * _get_int_0(n, prod, asum)
                        + 0.25 * _get_int_0(n, prod, bsum)
                    )
                else:
                    l0mats.append(
                        num * _get_int_d(n, prod, isum)
                        + 0.25 * _get_int_d(n, prod, asum)
                        + 0.25 * _get_int_d(n, prod, bsum)
                    )
                # print(ratio, r, n, rdr, np.linalg.cond(l0mats[-1]))
            for n, rdr in self.settings.iterate_l1_terms(ratio):
                # num = 0.25 * (ratio ** (0.5 * n) + ratio ** (-0.5 * n))
                num = 0.5
                if not rdr:
                    l1mats.append(
                        num * _get_int_1(n, prod, isum)
                        + 0.25 * _get_int_1(n, prod, asum)
                        + 0.25 * _get_int_1(n, prod, bsum)
                    )
                else:
                    l1mats.append(
                        num * _get_int_1d(n, prod, isum)
                        + 0.25 * _get_int_1d(n, prod, asum)
                        + 0.25 * _get_int_1d(n, prod, bsum)
                    )
                # print(ratio, r, n, rdr, np.linalg.cond(l1mats[-1]))
        self._num_l0_feat = len(l0mats)
        self._num_l1_feat = len(l1mats)
        coul_list = [m for m in l0mats + l1mats]
        LJ_list = [cholesky(coul, lower=True) for coul in coul_list]
        # for coul, LJ in zip(coul_list, LJ_list):
        #     print("HI", np.linalg.norm(coul - LJ.dot(LJ.T)))
        r2vals = 2.0 / self.alphas
        vals = np.exp(-self.alphas[None, :] * r2vals[:, None])
        vals *= self.alpha_norms[None, :]
        # print(np.linalg.cond(vals))
        self.fit_matrices = [
            np.ascontiguousarray(np.linalg.solve(vals.T, LJ).T) for LJ in LJ_list
        ]
        for coul, LJ in zip(coul_list, self.fit_matrices):
            tmp = np.linalg.solve(vals.T, coul).T
            tmp = np.linalg.solve(vals.T, tmp)
            # print(0, np.linalg.norm(tmp - tmp.T))
            # print(1, np.linalg.norm(tmp - LJ.T.dot(LJ)))
        self.alpha_norms = 4 * self.alphas**1.5 / ((4 - np.sqrt(2)) * np.pi**1.5)

    @property
    def num_l0_feat(self):
        return self._num_l0_feat

    @property
    def num_l1_feat(self):
        return self._num_l1_feat


class SDMXIntPlan(SDMXBasePlan):
    def __init__(
        self,
        settings,
        nspin,
        alpha0,
        lambd,
        nalpha,
    ):
        """

        Args:
            settings (SDMXFullSettings):
            nspin (int):
            alpha0 (float):
            lambd (float):
            nalpha (int):
        """
        self.settings = settings
        self.nspin = nspin
        if alpha0 <= 0:
            raise ValueError("alpha0 must be positive")
        self.alpha0 = alpha0
        if lambd <= 1:
            raise ValueError("lambd must be > 1")
        self.lambd = lambd
        if not isinstance(nalpha, int) or nalpha <= 0:
            raise ValueError("nalpha must be positive integer")
        self.nalpha = nalpha
        self.alphas = self.alpha0 * self.lambd ** np.arange(self.nalpha)
        self.alpha_norms = (np.pi / (2 * self.alphas)) ** -0.75
        r2vals = 2.0 / self.alphas
        vals = np.exp(-2 * self.alphas[None, :] * r2vals[:, None])
        vals *= self.alpha_norms[None, :] * self.alpha_norms[None, :]
        all_n = []
        self._num_l0_feat = 0
        self._num_l1_feat = 0
        for ratio in self.settings.ratios:
            for n, rdr in self.settings.iterate_l0_terms(ratio):
                if False:  # rdr or ratio != 1:
                    raise NotImplementedError(
                        "Numerical integral SDMX only works for ratio-1, rdr=False"
                    )
                all_n.append(n)
                self._num_l0_feat += 1
            for n, rdr in self.settings.iterate_l1_terms(ratio):
                if False:  # rdr or ratio != 1:
                    raise NotImplementedError(
                        "Numerical integral SDMX only works for ratio-1, rdr=False"
                    )
                all_n.append(n - 2)
                self._num_l1_feat += 1
        wt_dict = []
        for n in all_n:
            refz = True
            alphas = self.alphas
            if refz:
                alphas = np.append(alphas, [alphas[-1] * self.lambd])
            lambd = self.lambd
            mypow = 0.5 + 0.5 / lambd
            # mypow = 0.5
            amid = alphas[alphas.size // 2]
            rvals = (0.25 / mypow**2 / np.sqrt(amid)) * (amid / alphas) ** mypow
            # rvals = 1.0 / alphas**mypow
            if refz:
                rvals_abs = rvals.copy()
                rvals -= np.min(rvals)
            assert np.min(rvals) == rvals[-1]  # check ordering
            rdr = 4 * np.pi * rvals ** (2 - n) * rvals_abs * np.log(lambd**mypow)
            if refz:
                rdr = rdr[:-1]
                rvals = rvals[:-1]
            wt_dict.append(rdr)
            # a = self.alphas
            # prod = a**2
            # isum = 2 * a
            # ints = _get_int_0(n, prod, isum)
            # wt_dict.append(np.linalg.solve(vals.T, ints))
            # print(wt_dict[n])
        self.alphas = 2.0 / (rvals * rvals)
        self.wt_dict = wt_dict
        self.fit_metric = "ovlp"
        self.alpha_norms = 4 * self.alphas**1.5 / ((4 - np.sqrt(2)) * np.pi**1.5)

    @property
    def num_l0_feat(self):
        return self._num_l0_feat

    @property
    def num_l1_feat(self):
        return self._num_l1_feat

    def get_features(self, p_vag, out=None, l0tmp=None, l1tmp=None):
        """
        Should pass out[idm], l0tmp[idm], etc. to this function
        """
        na, ng = p_vag.shape[1:]
        nfeat = self.settings.nfeat
        n0 = self.num_l0_feat
        n1 = self.num_l1_feat
        if out is None:
            out = np.empty((nfeat, ng))
        if l0tmp is None:
            l0tmp = np.empty((na, ng))
        elif l1tmp is None:
            l1tmp = np.empty((3, na, ng))
        fac = -0.25 * self.nspin * self.nspin
        l0tmp[:] = p_vag[0]
        if l0tmp.dtype == np.float64:
            p0_ag = l0tmp * l0tmp
        else:
            p0_ag = l0tmp.real * l0tmp.real + l0tmp.imag * l0tmp.imag
        if n1 > 0:
            for v in range(3):
                l1tmp[v] = p_vag[v + 1]
            p1_ag = _sdmx_einsum(l1tmp, mystr="vag,vag->ag")
            p1_ag = np.einsum("vag,vag->ag", l1tmp, l1tmp)
        for ifeat in range(n0):
            out[ifeat] = pyscflib.einsum("ag,a->g", p0_ag, self.wt_dict[ifeat])
            out[ifeat] *= fac
        for ifeat in range(n0, n0 + n1):
            out[ifeat] = pyscflib.einsum("ag,a->g", p1_ag, self.wt_dict[ifeat])
            out[ifeat] *= fac
        return out

    def get_vxc(self, vxc_ig, l0tmp, l1tmp=None, out=None):
        nfeat = self.settings.nfeat
        assert nfeat == vxc_ig.shape[0]
        ng = vxc_ig.shape[1]
        n0 = self.num_l0_feat
        n1 = self.num_l1_feat
        if n1 == 0:
            nv = 1
        else:
            nv = 4
            assert l1tmp is not None
        if out is None:
            out = np.zeros((nv, self.nalpha, ng), dtype=l0tmp.dtype)
        else:
            out[:] = 0.0
        if nfeat == 0:
            return out
        fac = -0.25 * self.nspin * self.nspin
        v_ag = np.zeros((self.nalpha, ng))
        for ifeat in range(n0):
            v_ag[:] += self.wt_dict[ifeat][:, None] * vxc_ig[ifeat]
        out[0] = fac * l0tmp * v_ag
        if n1 > 0:
            v_ag[:] = 0
            for ifeat in range(n0, n0 + n1):
                v_ag[:] += self.wt_dict[ifeat][:, None] * vxc_ig[ifeat]
            for v in range(3):
                out[v + 1] = fac * l1tmp[v] * v_ag
        return out


class HybridPlan:
    """Plan for hybrid DFT. Not yet implemented"""

    def __init__(self, settings, nspin):
        """
        Args:
            settings (HybridSettings)
        """
        self.settings = settings
        self.nspin = nspin


class NLDFAuxiliaryPlan(ABC):
    """
    Plan for expanding the NLDFs using an auxiliary basis or spline.
    Stores the list of interpolation exponents and how they were constructed,
    as well as instructions for evaluating the auxiliary coefficients.
    Also has a method to call the relevant C function to get the coefficients.
    """

    def __init__(
        self,
        nldf_settings,
        nspin,
        alpha0,
        lambd,
        nalpha,
        coef_order="gq",
        alpha_formula="etb",
        proc_inds=None,
        rhocut=1e-10,
        expcut=1e-10,
    ):
        """
        Initialize NLDFAuxiliaryPlan

        Args:
            nldf_settings (NLDFSettings): code-universal NLDF settings
            nspin (int): 1 (non-spin-polarized) or 2 (spin-polarized)
            alpha0 (float): Small-exponent setting for constructing
                list of alphas.
            lambd (float): Inverse density of interpolation.
                Must be > 1. 1=infinitesimal spacing of Gaussian exponents
            nalpha (int): Number of interpolating exponents
            coef_order (str): Whether interpolation/expansion coefficients
                have grid as the slow (first) index ('gq') or exponent ('qg')
            alpha_formula (str): How to generate the interpolating exponents.
                Must be 'etb' (alphas[j] = alpha0 * lambd**j))
                or 'zexp' (alphas[j] = alpha0 * (lambd**j - 1) / (lambd - 1))
            proc_inds (list or np.ndarray): If parallelized, which
                alpha indexes does this process cover. List or array
                of integers.
            rhocut (float): Small-density cutoff for stability
            expcut (float): Small-exponent cutoff for stability
        """
        if not isinstance(nldf_settings, NLDFSettings):
            raise ValueError("Require NLDFSettings object")
        self.nldf_settings = nldf_settings
        if alpha0 <= 0:
            raise ValueError("alpha0 must be positive")
        self.alpha0 = alpha0
        if lambd <= 1:
            raise ValueError("lambd must be > 1")
        self.lambd = lambd
        if not isinstance(nalpha, int) or nalpha <= 0:
            raise ValueError("nalpha must be positive integer")
        self.nalpha = nalpha
        if alpha_formula == "zexp":
            self.alpha0 /= self.lambd - 1
        elif alpha_formula != "etb":
            raise ValueError
        self.alpha_formula = alpha_formula
        if coef_order not in ["gq", "qg"]:
            raise ValueError
        self.coef_order = coef_order
        # if alpha_order not in ['increasing', 'decreasing']:
        #    raise ValueError
        # if alpha_order == 'decreasing':
        #    raise NotImplementedError
        # one could hypothetically set these two differently,
        # but probably not useful to do so
        self.alpha_order = "increasing"
        self.normalization = "norm"

        # set the parallelization setting (which indices on this process)
        self.proc_inds = proc_inds

        if nspin not in [1, 2]:
            raise ValueError("nspin must be 1 (non-polarized) or 2 (polarized)")
        self.nspin = nspin
        if rhocut < 0 or expcut < 0:
            raise ValueError("rhocut and expcut must be nonnegative")
        self.rhocut = rhocut / nspin
        self.expcut = expcut

        self.alphas = None
        self.amin = None
        self.local_alphas = None
        self.local_nalpha = None
        self.alpha_norms = None
        self._alpha_transform = None
        self._local_qloc = None
        self._global_qloc = None

        self._l0_start_locs = []
        self._l1_start_locs = []
        self._cached_l1_data = {s: [] for s in range(self.nspin)}
        self._l1_dots = []
        self._cached_p_i_qg = {s: [] for s in range(self.nspin)}

        # set up the indexing helper for version i features
        start_locs = []
        end_locs = []
        loc_curr = 0
        for i in range(len(self.nldf_settings.l0_feat_specs)):
            start_locs.append(loc_curr)
            loc_curr += 1
            end_locs.append(loc_curr)
        self._l0_start_locs = start_locs
        start_locs = []
        end_locs = []
        for i in range(len(self.nldf_settings.l1_feat_specs)):
            start_locs.append(loc_curr)
            loc_curr += 3
            end_locs.append(loc_curr)
        self._l1_start_locs = start_locs
        self._l1_dots = self.nldf_settings.l1_feat_dots

        # set up the list of interpolating exponents
        # and their normalization coefficients
        self.alphas = self.get_q2a(np.arange(self.nalpha))
        if self.alpha_order == "decreasing":
            self.alphas = np.ascontiguousarray(np.flip(self.alphas))
        if self.proc_inds is not None:
            self.local_alphas = np.ascontiguousarray(self.alphas[self.proc_inds])
        else:
            self.local_alphas = self.alphas.copy()
        self.local_nalpha = self.local_alphas.size

        if self.normalization == "one":
            self.alpha_norms = np.ones_like(self.alphas)
        else:
            self.alpha_norms = (np.pi / (2 * self.alphas)) ** -0.75

        self._run_setup()

    def new(self, **kwargs):
        new_kwargs = dict(
            nldf_settings=self.nldf_settings,
            nspin=self.nspin,
            alpha0=self.alpha0,
            lambd=self.lambd,
            nalpha=self.nalpha,
            coef_order=self.coef_order,
            alpha_formula=self.alpha_formula,
            proc_inds=self.proc_inds,
            rhocut=self.rhocut,
            expcut=self.expcut,
        )
        new_kwargs.update(kwargs)
        return self.__class__(**new_kwargs)

    @abstractmethod
    def _run_setup(self):
        pass

    @property
    def num_vj(self):
        """Number of vj or vk features"""
        return self.nldf_settings.num_feat_param_sets

    @property
    def num_vi_ints(self):
        return len(self._l0_start_locs) + 3 * len(self._l1_start_locs)

    def get_q2a(self, q):
        """Convert (float) q index to exponent alpha."""
        if self.alpha_formula == "etb":
            return self.alpha0 * self.lambd**q
        else:
            res = self.alpha0 * (self.lambd**q - 1)
            res[0] = self.expcut
            return res

    def _get_extra_args(self, i):
        assert 0 <= i < self.nldf_settings.num_feat_param_sets
        params = self.nldf_settings.feat_params[i]
        # return [] if len(params) < 4 else params[3:]
        # TODO this works for GGA and MGGA but not if there are
        # multiple extra args for the feature specs. This might
        # need to be revised later on.
        if self.nldf_settings.feat_spec_list[i] == "se_erf_rinv":
            return [params[-1]]
        else:
            return []

    def get_function_to_convolve(self, rho_tuple):
        """
        Returns the function that gets convolved to make the nonlocal
        density features. Usually the density rho (settings.rho_mult='one'),
        but can also be other quantities like the rho times the theta
        exponent (rho_mult='expnt').

        Args:
            rho (np.ndarray): Density
            sigma (np.ndarray): Squared gradient of density
            tau (np.ndarray): Kinetic energy density

        Returns:
            tuple(np.ndarray): function to convolve, as well as its derivatives
            with respect to rho, sigma, and tau. Combined, this makes for a
            tuple of four arrays.
        """
        rho = rho_tuple[0]
        if self.nldf_settings.rho_mult == "one":
            da = [
                np.ones_like(rho),
            ] + [np.zeros_like(r) for r in rho_tuple[1:]]
            return rho, tuple(da)
        elif self.nldf_settings.rho_mult == "expnt":
            a, da_tuple = self.eval_feat_exp(rho_tuple, i=-1)
            for da in da_tuple:
                da[:] *= rho
            da_tuple[0][:] += a
            a[:] *= rho
            return a, da_tuple
        else:
            raise NotImplementedError

    def get_rho_tuple(self, rho_data, with_spin=False):
        return get_rho_tuple(
            rho_data, with_spin=with_spin, is_mgga=self.nldf_settings.sl_level == "MGGA"
        )

    def get_drhodf_tuple(self, rho_data, drhodf_data, with_spin=False):
        return get_drhodf_tuple(
            rho_data,
            drhodf_data,
            with_spin=with_spin,
            is_mgga=self.nldf_settings.sl_level == "MGGA",
        )

    def eval_feat_exp(self, rho_tuple, i=-1):
        """
        Evaluate the exponents that determine the length-scale
        of feature i at each grid point. i can take a range
        from -1 (theta exponent) to the number of version j/k
        features minus 1.

        Args:F
            rho (np.ndarray): density
            sigma (np.ndarray): squared gradient
            tau (np.ndarray): kinetic energy density
            i: feature id. -1 corresponds to theta exponent

        Returns:
            the cider exponent array (np.ndarray)
        """
        if self.nldf_settings.sl_level == "MGGA":
            rho, sigma, tau = rho_tuple
            if i == -1:
                a0, grad_mul, tau_mul = self.nldf_settings.theta_params
            else:
                if not 0 <= i < self.nldf_settings.num_feat_param_sets:
                    raise ValueError("Feature index out of range")
                a0, grad_mul, tau_mul = self.nldf_settings.feat_params[i][:3]
            a, dadn, dadsigma, dadtau = get_cider_exponent(
                rho,
                sigma,
                tau,
                a0=a0,
                grad_mul=grad_mul,
                tau_mul=tau_mul,
                rhocut=self.rhocut,
                nspin=self.nspin,
            )
            return a, (dadn, dadsigma, dadtau)
        else:
            rho, sigma = rho_tuple
            if i == -1:
                a0, grad_mul = self.nldf_settings.theta_params[:2]
            else:
                if not 0 <= i < self.nldf_settings.num_feat_param_sets:
                    raise ValueError("Feature index out of range")
                a0, grad_mul = self.nldf_settings.feat_params[i][:2]
            a, dadn, dadsigma = get_cider_exponent_gga(
                rho,
                sigma,
                a0=a0,
                grad_mul=grad_mul,
                rhocut=self.rhocut,
                nspin=self.nspin,
            )
            return a, (dadn, dadsigma)

    def _clear_l1_cache(self, s):
        self._cached_l1_data[s] = []

    def _clear_p_cache(self, s):
        self._cached_p_i_qg[s] = []

    def _cache_l1_vectors(self, f_qg, drho, s):
        for loc in self._l1_start_locs:
            self._cached_l1_data[s].append(f_qg[loc : loc + 3].copy())
        self._cached_l1_data[s].append(drho.copy())

    def _cache_p_tensor(self, s, p):
        self._cached_p_i_qg[s].append(p)

    def eval_rho_vj_(self, f_qg, p_i_qg, feat):
        if len(p_i_qg) == 0:
            return feat
        if self.coef_order == "gq":
            p_i_qg = [p.T for p in p_i_qg]
        nalpha = p_i_qg[0].shape[0]
        for i in range(len(p_i_qg)):
            feat[i] = np.einsum("qg,qg->g", f_qg[:nalpha], p_i_qg[i])
        return feat

    def eval_vxc_vj_(self, vfeat, p_i_qg, vf_qg):
        if len(p_i_qg) == 0:
            return vf_qg
        if self.coef_order == "gq":
            p_i_qg = [p.T for p in p_i_qg]
        nalpha = p_i_qg[0].shape[1]
        for i in range(len(p_i_qg)):
            vf_qg[:nalpha] += p_i_qg[i] * vfeat[i]
        return vf_qg

    def eval_rho_vi_(self, f_qg, drho, feat, spin):
        self._clear_l1_cache(spin)
        self._cache_l1_vectors(f_qg, drho, spin)
        i = 0
        l1cache = self._cached_l1_data[spin]
        for loc in self._l0_start_locs:
            feat[i] = f_qg[loc]
            i += 1
        for j, k in self._l1_dots:
            feat[i] = np.einsum("xg,xg->g", l1cache[j], l1cache[k])
            feat[i] *= self.nspin
            i += 1
        return feat

    def eval_vxc_vi_(self, vfeat, vdrho, vf_qg, spin):
        if len(self._cached_l1_data[spin]) != len(self._l1_start_locs) + 1:
            raise ValueError("Must have cached l=1 features to get potential")
        l1cache = self._cached_l1_data[spin]
        i = 0
        for loc in self._l0_start_locs:
            vf_qg[loc] = vfeat[i]
            i += 1
        for j, k in self._l1_dots:
            vfeat[i] *= self.nspin
            if j == -1:
                target = vdrho
            else:
                loc = self._l1_start_locs[j]
                target = vf_qg[loc : loc + 3]
            target[:] += vfeat[i] * l1cache[k]
            if k == -1:
                target = vdrho
            else:
                loc = self._l1_start_locs[k]
                target = vf_qg[loc : loc + 3]
            target[:] += vfeat[i] * l1cache[j]
            i += 1
        # self._clear_l1_cache(spin)

    def eval_rho_full(
        self,
        f,
        rho_data,
        spin=0,
        feat=None,
        dfeat=None,
        cache_p=True,
        apply_transformation=False,
        coeff_multipliers=None,
    ):
        """

        Args:
            f:
            rho: [n, dn/dx, dn/dy, dn/dz, (tau)]
            drho:
            tau:
            feat:
            dfeat:
            cache_p:

        Returns:
            If apply_transformation is True, f is overwritten
        """
        if self.coef_order == "qg":
            f_qg = f
        else:
            f_qg = f.T
        ngrids = f_qg.shape[1]
        rho_tuple = self.get_rho_tuple(rho_data)
        num_vj = self.nldf_settings.num_feat_param_sets
        if feat is None:
            feat = np.empty((self.nldf_settings.nfeat, f_qg.shape[1]))
        if dfeat is None:
            dfeat = np.empty((num_vj, f_qg.shape[1]))
        self._clear_p_cache(spin)
        dbuf = self.empty_coefs(ngrids, local=False)
        for i in range(num_vj):
            a_g = self.get_interpolation_arguments(rho_tuple, i=i)[0]
            p, dp = self.get_interpolation_coefficients(a_g, i=i, dbuf=dbuf)
            if coeff_multipliers is not None:
                p[:] *= coeff_multipliers[:, None]
                dp[:] *= coeff_multipliers[:, None]
            if apply_transformation:
                self.get_transformed_interpolation_terms(p, i=i, fwd=True, inplace=True)
                self.get_transformed_interpolation_terms(
                    dp, i=i, fwd=True, inplace=True
                )
            self.eval_rho_vj_(f_qg[: self.nalpha], [dp], dfeat[i : i + 1])
            self.eval_rho_vj_(f_qg[: self.nalpha], [p], feat[i : i + 1])
            if cache_p:
                self._cache_p_tensor(spin, p)
        start = 0 if self.nldf_settings.nldf_type == "i" else self.nalpha
        self._clear_l1_cache(spin)
        if "i" in self.nldf_settings.nldf_type:
            self.eval_rho_vi_(f_qg[start:], rho_data[1:4], feat[num_vj:], spin=spin)
        feat[:] *= self.nspin
        # dfeat[:] *= self.nspin
        return feat, dfeat

    def eval_occd_full(
        self,
        f,
        rho_data,
        occd_f,
        occd_rho_data,
        apply_transformation=True,
        coeff_multipliers=None,
    ):
        spin = 0
        # if self.nspin != 1:
        #     raise NotImplementedError
        if self.coef_order == "qg":
            f_qg = f
            occd_f_qg = occd_f
        else:
            f_qg = f.T
            occd_f_qg = occd_f.T
        ngrids = f_qg.shape[1]
        rho_tuple = self.get_rho_tuple(rho_data)
        occd_sigma = 2 * np.einsum("xg,xg->g", rho_data[1:4], occd_rho_data[1:4])
        num_vj = self.nldf_settings.num_feat_param_sets
        occd1_feat = np.zeros((self.nldf_settings.nfeat, f_qg.shape[1]))
        occd2_feat = np.zeros((self.nldf_settings.nfeat, f_qg.shape[1]))
        dbuf = self.empty_coefs(ngrids, local=False)
        for i in range(num_vj):
            a_g, da_tuple = self.get_interpolation_arguments(rho_tuple, i=i)
            p, dp = self.get_interpolation_coefficients(a_g, i=i, dbuf=dbuf)
            if coeff_multipliers is not None:
                p[:] *= coeff_multipliers[:, None]
                dp[:] *= coeff_multipliers[:, None]
            occd_a_g = occd_rho_data[0] * da_tuple[0]
            occd_a_g[:] += occd_sigma * da_tuple[1]
            if self.nldf_settings.sl_level == "MGGA":
                occd_a_g[:] += occd_rho_data[4] * da_tuple[2]
            if self.coef_order == "qg":
                dp *= occd_a_g
            else:
                dp *= occd_a_g[:, None]
            if apply_transformation:
                self.get_transformed_interpolation_terms(p, i=i, fwd=True, inplace=True)
                self.get_transformed_interpolation_terms(
                    dp, i=i, fwd=True, inplace=True
                )
            self.eval_rho_vj_(f_qg[: self.nalpha], [dp], occd1_feat[i : i + 1])
            self.eval_rho_vj_(occd_f_qg[: self.nalpha], [p], occd2_feat[i : i + 1])
        start = 0 if self.nldf_settings.nldf_type == "i" else self.nalpha
        self._clear_l1_cache(spin)
        self._cache_l1_vectors(f_qg[start:], rho_data[1:4], spin)
        l1_vals = [v for v in self._cached_l1_data[spin]]
        self._clear_l1_cache(spin)
        self._cache_l1_vectors(occd_f_qg[start:], occd_rho_data[1:4], spin)
        l1_occd = [v for v in self._cached_l1_data[spin]]
        i = 0
        for loc in self._l0_start_locs:
            occd1_feat[num_vj + i] = occd_f_qg[start + loc]
            i += 1
        for j, k in self._l1_dots:
            occd1_feat[num_vj + i] = np.einsum(
                "xg,xg->g", l1_vals[j], l1_occd[k]
            ) + np.einsum("xg,xg->g", l1_occd[j], l1_vals[k])
            i += 1
        return self.nspin * (occd1_feat + occd2_feat)

    def eval_vxc_full(
        self,
        vfeat,
        vrho_data,
        dfeat,
        rho_data,
        spin=0,
        p_i_qg=None,
        vf=None,
    ):
        """

        Args:
            vfeat:
            vrho:
            vdrho:
            vtau:
            dfeat:
            rho:
            drho:
            tau:
            drho:
            p_i_qg:
            vf:

        Returns:
            vf (nalpha + num_vi_ints + vf_buffer_nslot, ngrids)
                    or transpose if coef_order == 'gq':
                The last vf_buffer_nslot rows (qg) or columns (gq)
                are zeros. The remaining rows are filled with the
                functional derivatives with respect to the nonlocal
                density integrals.
        """
        vfeat[:] *= self.nspin
        if vf is None:
            vf = self.zero_coefs_full(vfeat.shape[1])
        if self.coef_order == "qg":
            vf_qg = vf
        else:
            vf_qg = vf.T
        num_vj = self.nldf_settings.num_feat_param_sets
        if p_i_qg is None:
            assert len(self._cached_p_i_qg[spin]) == num_vj
            p_i_qg = self._cached_p_i_qg[spin]
        if self.nldf_settings.nldf_type == "i":
            start = 0
        else:
            start = self.nalpha
            self.eval_vxc_vj_(vfeat[:num_vj], p_i_qg, vf_qg[:start])
        if "i" in self.nldf_settings.nldf_type:
            self.eval_vxc_vi_(vfeat[num_vj:], vrho_data[1:4], vf_qg[start:], spin)
        num_vj = self.nldf_settings.num_feat_param_sets
        rho_tuple = self.get_rho_tuple(rho_data)
        for i in range(num_vj):
            da_tuple = self.get_interpolation_arguments(rho_tuple, i=i)[1]
            vrho_data[0, :] += da_tuple[0] * dfeat[i] * vfeat[i]
            vrho_data[1:4, :] += 2 * rho_data[1:4] * da_tuple[1] * dfeat[i] * vfeat[i]
            if self.nldf_settings.sl_level == "MGGA":
                vrho_data[4, :] += da_tuple[2] * dfeat[i] * vfeat[i]
        # self._clear_p_cache(spin)
        return vf_qg if self.coef_order == "qg" else vf_qg.T

    @abstractmethod
    def _get_interpolation_arguments(self, rho_tuple, i=-1):
        pass

    @abstractmethod
    def _get_interpolation_coefficients(self, arg_g, i=-1, vbuf=None, dbuf=None):
        pass

    @abstractmethod
    def _get_transformed_interpolation_terms(self, p_xx, i=-1, fwd=True, inplace=False):
        pass

    def get_interpolation_arguments(self, rho_tuple, i=-1):
        if i == -1 and self.nldf_settings.nldf_type == "k":
            return self.eval_feat_exp(rho_tuple, i=i)
        else:
            return self._get_interpolation_arguments(rho_tuple, i=i)

    def get_interpolation_coefficients(self, arg_g, i=-1, vbuf=None, dbuf=None):
        if not (-1 <= i < self.nldf_settings.num_feat_param_sets):
            raise ValueError("Unsupported feature index")
        if i == -1 and self.nldf_settings.nldf_type == "k":
            ngrids = arg_g.size
            nalpha = self.local_nalpha
            if self.coef_order == "gq":
                fn = libcider.cider_coefs_vk1_gq
                shape = (ngrids, nalpha)
            else:
                fn = libcider.cider_coefs_vk1_qg
                shape = (nalpha, ngrids)
            p = np.ndarray(shape, dtype=np.float64, buffer=vbuf)
            dp = np.ndarray(shape, dtype=np.float64, buffer=dbuf)
            fn(
                p.ctypes.data_as(ctypes.c_void_p),
                dp.ctypes.data_as(ctypes.c_void_p),
                arg_g.ctypes.data_as(ctypes.c_void_p),
                self.local_alphas.ctypes.data_as(ctypes.c_void_p),
                ctypes.c_int(arg_g.size),
                ctypes.c_int(self.nalpha),
            )
            return p, dp
        else:
            return self._get_interpolation_coefficients(
                arg_g, i=i, vbuf=vbuf, dbuf=dbuf
            )

    def get_transformed_interpolation_terms(self, p_xx, i=-1, fwd=True, inplace=False):
        if i == -1 and self.nldf_settings.nldf_type == "k":
            return p_xx if inplace else p_xx.copy()
        else:
            return self._get_transformed_interpolation_terms(
                p_xx, i=i, fwd=fwd, inplace=inplace
            )

    def _get_coef_shape(self, ngrids, local, nextra=0):
        nalpha = self.local_nalpha if local else self.nalpha
        nalpha += nextra
        if self.coef_order == "gq":
            return (ngrids, nalpha)
        else:
            return (nalpha, ngrids)

    def empty_coefs(self, ngrids, local=True, buf=None):
        shape = self._get_coef_shape(ngrids, local)
        return np.ndarray(shape, dtype=np.float64, order="C", buffer=buf)

    def zero_coefs(self, ngrids, local=True):
        shape = self._get_coef_shape(ngrids, local)
        return np.zeros(shape, dtype=np.float64, order="C")

    def zero_coefs_full(self, ngrids, buffer_nrows=0):
        nextra = buffer_nrows + self.num_vi_ints
        if self.nldf_settings.nldf_type == "i":
            if self.coef_order == "gq":
                shape = (ngrids, nextra)
            else:
                shape = (nextra, ngrids)
        else:
            shape = self._get_coef_shape(ngrids, False, nextra=nextra)
        return np.zeros(shape, dtype=np.float64, order="C")


class NLDFGaussianPlan(NLDFAuxiliaryPlan):
    def _run_setup(self):
        # alpha_sinv, norm, transform = get_diff_gaussian_ovlp(
        #    self.alphas, self.alpha_norms,
        # )[1:4]
        # self._alpha_transform = transform.T.dot(alpha_sinv)
        # self._alpha_transform /= np.sqrt(norm)
        ovlp, _ = _get_ovlp_fit_interpolation_coefficients(
            self, self.alphas, i=-1, local=False
        )
        ovlp *= self.alpha_norms[None, :] * self.alpha_norms[:, None]
        self._dmul = False
        if self._dmul:
            ref = np.identity(ovlp.shape[0]) * self.alpha_norms[:, None]
            self._alpha_transform = _stable_solve(ovlp, ref)
        else:
            self._alpha_transform = ovlp

    def _get_interpolation_arguments(self, rho_tuple, i=-1):
        return self.eval_feat_exp(rho_tuple, i=i)

    def _get_interpolation_coefficients(
        self, arg_g, i=-1, local=True, vbuf=None, dbuf=None
    ):
        p, dp = _get_ovlp_fit_interpolation_coefficients(
            self,
            arg_g,
            i=i,
            local=local,
            vbuf=vbuf,
            dbuf=dbuf,
        )
        # p[1:] -= p[:-1]
        # dp[1:] -= dp[:-1]
        return p, dp

    def _get_transformed_interpolation_terms(self, p_xx, i=-1, fwd=True, inplace=False):
        if self.coef_order == "gq":
            in_size = p_xx.shape[1]
        elif self.coef_order == "qg":
            in_size = p_xx.shape[0]
        else:
            raise ValueError
        if in_size != self.nalpha:
            raise ValueError("Wrong number of interpolating points")
        if fwd:
            transform = self._alpha_transform
        else:
            transform = self._alpha_transform.T
        if self.coef_order == "gq":
            p_qu = p_xx.transpose()
        else:
            p_qu = p_xx
        if self._dmul:
            # TODO parallelize with openmp
            if inplace:
                p_qu[:] = transform.dot(p_qu)
            else:
                p_qu = transform.dot(p_qu)
        else:
            if inplace:
                if fwd:
                    p_qu[:] *= self.alpha_norms[:, None]
                p_qu[:] = _stable_solve(transform, p_qu)
                if not fwd:
                    p_qu[:] *= self.alpha_norms[:, None]
            else:
                if fwd:
                    p_qu = self.alpha_norms[:, None] * p_qu
                p_qu = _stable_solve(transform, p_qu)
                if not fwd:
                    p_qu = self.alpha_norms[:, None] * p_qu
        if self.coef_order == "gq":
            p_xx = p_qu.T
        else:
            p_xx = p_qu
        return p_xx


def _stable_solve(a, b):
    c_and_l = cho_factor(a)
    return cho_solve(c_and_l, b)


class NLDFSplinePlan(NLDFAuxiliaryPlan):
    def __init__(
        self,
        nldf_settings,
        nspin,
        alpha0,
        lambd,
        nalpha,
        coef_order="gq",
        alpha_formula="etb",
        proc_inds=None,
        spline_size=None,
        rhocut=1e-10,
        expcut=1e-10,
    ):
        self._spline_size = nalpha if spline_size is None else spline_size
        self._local_alpha_transform = None
        super(NLDFSplinePlan, self).__init__(
            nldf_settings,
            nspin,
            alpha0,
            lambd,
            nalpha,
            coef_order=coef_order,
            alpha_formula=alpha_formula,
            proc_inds=proc_inds,
            rhocut=rhocut,
            expcut=expcut,
        )

    def _run_setup(self):
        ovlp, _ = _get_ovlp_fit_interpolation_coefficients(
            self, self.alphas, i=-1, local=False
        )
        ovlp *= self.alpha_norms[None, :] * self.alpha_norms[:, None]
        self._alpha_transform = []
        self._local_alpha_transform = []
        # interp_indexes = np.linspace(0, self.nalpha - 1, self._spline_size)
        interp_indexes = np.arange(0, self._spline_size).astype(np.float64)
        dense_alphas = self.get_q2a(
            interp_indexes * (self.nalpha - 1) / (self._spline_size - 1)
        )
        for i in range(self.nldf_settings.num_feat_param_sets):
            p, _ = _get_ovlp_fit_interpolation_coefficients(
                self, dense_alphas, i=i, local=False
            )
            if self.coef_order == "gq":
                p = p.T
            p[:, :] *= self.alpha_norms[:, None]
            coefs_qi = np.ascontiguousarray(_stable_solve(ovlp, p))
            self._alpha_transform.append(
                _construct_cubic_splines(coefs_qi, interp_indexes)
            )
        p, _ = _get_ovlp_fit_interpolation_coefficients(
            self, dense_alphas, i=-1, local=False
        )
        if self.coef_order == "gq":
            p = p.T
        p[:, :] *= self.alpha_norms[:, None]
        coefs_qi = np.ascontiguousarray(_stable_solve(ovlp, p))
        self._alpha_transform.append(_construct_cubic_splines(coefs_qi, interp_indexes))
        self._local_alpha_transform = [
            np.ascontiguousarray(t[:, self.proc_inds, :]) for t in self._alpha_transform
        ]
        if self.alpha_formula == "zexp":
            self.alpha_norms[0] = 0  # for stability

    def get_a2q_fast(self, exp_g):
        """A fast (parallel C-backend) version of get_a2q that
        also has a local option to get the q values just
        for the local alphas."""
        # TODO add bounds checking
        di = np.empty_like(exp_g)
        derivi = np.empty_like(exp_g)
        if self.alpha_formula == "etb":
            fn = libcider.cider_ind_etb
        else:
            fn = libcider.cider_ind_zexp
        fn(
            di.ctypes.data_as(ctypes.c_void_p),
            derivi.ctypes.data_as(ctypes.c_void_p),
            exp_g.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(exp_g.size),
            ctypes.c_double(self.alpha0),
            ctypes.c_double(self.lambd),
        )
        # if self.alpha_order == 'decreasing':
        #    di[:] = self.nalpha - 1 - di
        #    derivi[:] *= -1
        if self._spline_size != self.nalpha:
            di[:] *= (self._spline_size - 1) / (self.nalpha - 1)
            derivi[:] *= (self._spline_size - 1) / (self.nalpha - 1)
        libcider.cider_ind_clip(
            di.ctypes.data_as(ctypes.c_void_p),
            derivi.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(self._spline_size - 1),
            ctypes.c_int(exp_g.size),
        )
        return di, derivi

    def _get_interpolation_arguments(self, rho_tuple, i=-1):
        a_g, da_tuple = self.eval_feat_exp(rho_tuple, i=i)
        di, derivi = self.get_a2q_fast(a_g)
        for da in da_tuple:
            da[:] *= derivi
        return di, da_tuple

    def _get_interpolation_coefficients(
        self, arg_g, i=-1, local=True, vbuf=None, dbuf=None
    ):
        ngrids = arg_g.size
        nalpha = self.local_nalpha if local else self.nalpha
        if self.coef_order == "gq":
            fn = libcider.cider_coefs_spline_gq
        else:
            fn = libcider.cider_coefs_spline_qg
        p = self.empty_coefs(ngrids, local=local, buf=vbuf)
        dp = self.empty_coefs(ngrids, local=local, buf=dbuf)
        if local:
            w_kap = self._local_alpha_transform[i]
        else:
            w_kap = self._alpha_transform[i]
        fn(
            p.ctypes.data_as(ctypes.c_void_p),
            dp.ctypes.data_as(ctypes.c_void_p),
            arg_g.ctypes.data_as(ctypes.c_void_p),
            w_kap.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(ngrids),
            ctypes.c_int(nalpha),
            ctypes.c_double(self.lambd),
        )
        return p, dp

    def _get_transformed_interpolation_terms(self, p_xx, i=-1, fwd=True, inplace=False):
        return p_xx if inplace else p_xx.copy()


def get_diff_gaussian_ovlp(alphas, alpha_norms, flip=False):
    potrf = scipy.linalg.lapack.get_lapack_funcs("potrf")
    trtri = scipy.linalg.lapack.get_lapack_funcs("trtri")

    alpha_s3 = (np.pi / (alphas + alphas[:, None])) ** 1.5
    alpha_s4 = alpha_s3.copy()
    alpha_s4[1:, 1:] += alpha_s3[:-1, :-1]
    alpha_s4[1:, :] -= alpha_s3[:-1, :]
    alpha_s4[:, 1:] -= alpha_s3[:, :-1]
    ad = np.diag(alpha_s4).copy()
    norm4 = ad
    alpha_s4 /= np.sqrt(ad * ad[:, None])
    alpha_s4 = alpha_s4
    transform = np.identity(alpha_s4.shape[0])
    for i in range(alpha_s4.shape[0] - 1):
        transform[i + 1, i] = -1
    transform /= np.sqrt(ad)[:, None]

    if flip:
        alpha_s4 = np.flip(alpha_s4, axis=0)
        alpha_s4 = np.flip(alpha_s4, axis=1)
    tmp, info = potrf(alpha_s4, lower=True)
    tmp, info = trtri(tmp, lower=True)
    tmp = np.tril(tmp)
    alpha_sinv4 = tmp.T.dot(tmp)
    if flip:
        alpha_sinv4 = np.flip(alpha_sinv4, axis=0)
        alpha_sinv4 = np.flip(alpha_sinv4, axis=1)
    transform = np.identity(tmp.shape[0])
    for i in range(tmp.shape[0] - 1):
        transform[i + 1, i] = -1
    transform /= alpha_norms
    transform /= np.sqrt(norm4[:, None])
    transform = transform

    return alpha_s4, alpha_sinv4, norm4, transform, transform.T.dot(tmp.T)



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import numpy as np
from gpaw.sphere.lebedev import R_nv, Y_nL, weight_n
from gpaw.xc.pawcorrection import rnablaY_nLv
from numpy import pi, sqrt

from ciderpress.gpaw.atom_utils import (
    CiderRadialExpansion,
    DiffPAWXCCorrection,
    FastPASDWCiderKernel,
    PAugSetup,
    PAWCiderContribs,
    PSmoothSetupV1,
    PSmoothSetupV2,
    SBTGridContainer,
    calculate_cider_paw_correction,
)
from ciderpress.gpaw.fit_paw_gauss_pot import get_dv


def get_kinetic_energy_sl(xcc, ae, D_sp):
    nspins = D_sp.shape[0]
    if ae:
        tau_pg = xcc.tau_pg
        tauc_g = xcc.tauc_g / (np.sqrt(4 * np.pi) * nspins)
    else:
        tau_pg = xcc.taut_pg
        tauc_g = xcc.tauct_g / (np.sqrt(4 * np.pi) * nspins)
    nn = tau_pg.shape[-1] // tauc_g.shape[0]
    tau_sg = np.dot(D_sp, tau_pg)
    tau_sg.shape = (tau_sg.shape[0], -1, nn)
    tau_sg[:] += tauc_g[:, None]
    tau_sg.shape = (tau_sg.shape[0], -1)
    return tau_sg


def vec_radial_vars_sl(xcc, n_sLg, Y_nL, dndr_sLg, rnablaY_nLv, ae, is_mgga, D_sp):
    nspin = len(n_sLg)
    ngrid = n_sLg.shape[-1] * weight_n.size
    nx = 5 if is_mgga else 4
    rho_sxg = np.empty((nspin, nx, ngrid))
    rho_sxg[:, 0] = np.dot(Y_nL, n_sLg).transpose(1, 2, 0).reshape(nspin, -1)
    b_vsgn = np.dot(rnablaY_nLv.transpose(0, 2, 1), n_sLg).transpose(1, 2, 3, 0)
    b_vsgn[..., 1:, :] /= xcc.rgd.r_g[1:, None]
    b_vsgn[..., 0, :] = b_vsgn[..., 1, :]
    a_sgn = np.dot(Y_nL, dndr_sLg).transpose(1, 2, 0)
    b_vsgn += R_nv.T[:, None, None, :] * a_sgn[None, :, :]
    N = Y_nL.shape[0]
    e_g = xcc.rgd.empty(N).reshape(-1)
    e_g[:] = 0
    rho_sxg[:, 1:4] = b_vsgn.transpose(1, 0, 2, 3).reshape(nspin, 3, -1)
    if is_mgga:
        rho_sxg[:, 4] = get_kinetic_energy_sl(xcc, ae, D_sp)
    vrho_sxg = np.zeros_like(rho_sxg)
    return e_g, rho_sxg, vrho_sxg


def get_kinetic_energy_and_occd(xcc, ae, D_sp, DD_op):
    nspins = D_sp.shape[0]
    if ae:
        tau_pg = xcc.tau_pg
        tauc_g = xcc.tauc_g / (np.sqrt(4 * np.pi) * nspins)
    else:
        tau_pg = xcc.taut_pg
        tauc_g = xcc.tauct_g / (np.sqrt(4 * np.pi) * nspins)
    nn = tau_pg.shape[-1] // tauc_g.shape[0]
    tau_sg = np.dot(D_sp, tau_pg)
    dtaudf_og = np.dot(DD_op, tau_pg)
    tau_sg.shape = (tau_sg.shape[0], -1, nn)
    tau_sg[:] += tauc_g[:, None]
    tau_sg.shape = (tau_sg.shape[0], -1)
    return tau_sg, dtaudf_og


def vec_radial_vars_occd(
    xcc,
    n_sLg,
    dndf_oLg,
    Y_nL,
    dndr_sLg,
    dndrdf_oLg,
    rnablaY_nLv,
    ae,
    D_sp,
    DD_op,
    is_mgga,
):
    nspin = len(n_sLg)
    norb = len(dndf_oLg)
    ngrid = n_sLg.shape[-1] * weight_n.size
    nx = 5 if is_mgga else 4

    rho_sxg = np.empty((nspin, nx, ngrid))
    drhodf_oxg = np.empty((norb, nx, ngrid))
    rho_sxg[:, 0] = np.dot(Y_nL, n_sLg).transpose(1, 2, 0).reshape(nspin, -1)
    drhodf_oxg[:, 0] = np.dot(Y_nL, dndf_oLg).transpose(1, 2, 0).reshape(norb, -1)

    b_vsgn = np.dot(rnablaY_nLv.transpose(0, 2, 1), n_sLg).transpose(1, 2, 3, 0)
    b_vsgn[..., 1:, :] /= xcc.rgd.r_g[1:, None]
    b_vsgn[..., 0, :] = b_vsgn[..., 1, :]
    a_sgn = np.dot(Y_nL, dndr_sLg).transpose(1, 2, 0)
    b_vsgn += R_nv.T[:, None, None, :] * a_sgn[None, :, :]
    rho_sxg[:, 1:4] = b_vsgn.transpose(1, 0, 2, 3).reshape(nspin, 3, -1)

    b_vogn = np.dot(rnablaY_nLv.transpose(0, 2, 1), dndf_oLg).transpose(1, 2, 3, 0)
    b_vogn[..., 1:, :] /= xcc.rgd.r_g[1:, None]
    b_vogn[..., 0, :] = b_vogn[..., 1, :]
    a_ogn = np.dot(Y_nL, dndrdf_oLg).transpose(1, 2, 0)
    b_vogn += R_nv.T[:, None, None, :] * a_ogn[None, :, :]
    drhodf_oxg[:, 1:4] = b_vogn.transpose(1, 0, 2, 3).reshape(norb, 3, -1)

    if is_mgga:
        rho_sxg[:, 4], drhodf_oxg[:, 4] = get_kinetic_energy_and_occd(
            xcc, ae, D_sp, DD_op
        )
    return rho_sxg, drhodf_oxg


def calculate_cider_paw_correction_deriv(
    expansion, setup, D_sp, DD_op=None, addcoredensity=True, has_cider_contribs=True
):
    xcc = setup.xc_correction
    rgd = xcc.rgd
    nspins = len(D_sp)

    if has_cider_contribs:
        setup.cider_contribs.set_D_sp(D_sp, setup.xc_correction)
        setup.cider_contribs._DD_op = DD_op

    if addcoredensity:
        nc0_sg = rgd.empty(nspins)
        nct0_sg = rgd.empty(nspins)
        dc0_sg = rgd.empty(nspins)
        dct0_sg = rgd.empty(nspins)
        nc0_sg[:] = sqrt(4 * pi) / nspins * xcc.nc_g
        nct0_sg[:] = sqrt(4 * pi) / nspins * xcc.nct_g
        dc0_sg[:] = sqrt(4 * pi) / nspins * xcc.dc_g
        dct0_sg[:] = sqrt(4 * pi) / nspins * xcc.dct_g
        if xcc.nc_corehole_g is not None and nspins == 2:
            nc0_sg[0] -= 0.5 * sqrt(4 * pi) * xcc.nc_corehole_g
            nc0_sg[1] += 0.5 * sqrt(4 * pi) * xcc.nc_corehole_g
            dc0_sg[0] -= 0.5 * sqrt(4 * pi) * xcc.dc_corehole_g
            dc0_sg[1] += 0.5 * sqrt(4 * pi) * xcc.dc_corehole_g
    else:
        nc0_sg = 0
        nct0_sg = 0
        dc0_sg = 0
        dct0_sg = 0

    D_sLq = np.inner(D_sp, xcc.B_pqL.T)
    DD_oLq = np.inner(DD_op, xcc.B_pqL.T)

    res = expansion(rgd, D_sLq, DD_oLq, xcc.n_qg, xcc.d_qg, nc0_sg, dc0_sg, ae=True)
    rest = expansion(
        rgd, D_sLq, DD_oLq, xcc.nt_qg, xcc.dt_qg, nct0_sg, dct0_sg, ae=False
    )

    return res, rest


class PAWCiderFeatContribs(PAWCiderContribs):
    def get_kinetic_energy_and_occd(self, ae):
        nspins = self._D_sp.shape[0]
        xcc = self.xcc
        if ae:
            tau_pg = xcc.tau_pg
            tauc_g = xcc.tauc_g / (np.sqrt(4 * np.pi) * nspins)
        else:
            tau_pg = xcc.taut_pg
            tauc_g = xcc.tauct_g / (np.sqrt(4 * np.pi) * nspins)
        nn = tau_pg.shape[-1] // tauc_g.shape[0]
        tau_sg = np.dot(self._D_sp, tau_pg)
        dtaudf_og = np.dot(self._DD_op, tau_pg)
        tau_sg.shape = (tau_sg.shape[0], -1, nn)
        tau_sg[:] += tauc_g[:, None]
        tau_sg.shape = (tau_sg.shape[0], -1)
        return tau_sg, dtaudf_og

    def vec_radial_vars_occd(
        self, n_sLg, dndf_oLg, Y_nL, dndr_sLg, dndrdf_oLg, rnablaY_nLv, ae
    ):
        nspin = len(n_sLg)
        norb = len(dndf_oLg)
        ngrid = n_sLg.shape[-1] * weight_n.size
        nx = 5 if self.is_mgga else 4

        rho_sxg = np.empty((nspin, nx, ngrid))
        drhodf_oxg = np.empty((norb, nx, ngrid))
        rho_sxg[:, 0] = np.dot(Y_nL, n_sLg).transpose(1, 2, 0).reshape(nspin, -1)
        drhodf_oxg[:, 0] = np.dot(Y_nL, dndf_oLg).transpose(1, 2, 0).reshape(norb, -1)

        b_vsgn = np.dot(rnablaY_nLv.transpose(0, 2, 1), n_sLg).transpose(1, 2, 3, 0)
        b_vsgn[..., 1:, :] /= self.xcc.rgd.r_g[1:, None]
        b_vsgn[..., 0, :] = b_vsgn[..., 1, :]
        a_sgn = np.dot(Y_nL, dndr_sLg).transpose(1, 2, 0)
        b_vsgn += R_nv.T[:, None, None, :] * a_sgn[None, :, :]
        rho_sxg[:, 1:4] = b_vsgn.transpose(1, 0, 2, 3).reshape(nspin, 3, -1)

        b_vogn = np.dot(rnablaY_nLv.transpose(0, 2, 1), dndf_oLg).transpose(1, 2, 3, 0)
        b_vogn[..., 1:, :] /= self.xcc.rgd.r_g[1:, None]
        b_vogn[..., 0, :] = b_vogn[..., 1, :]
        a_ogn = np.dot(Y_nL, dndrdf_oLg).transpose(1, 2, 0)
        b_vogn += R_nv.T[:, None, None, :] * a_ogn[None, :, :]
        drhodf_oxg[:, 1:4] = b_vogn.transpose(1, 0, 2, 3).reshape(norb, 3, -1)

        if self.is_mgga:
            rho_sxg[:, 4], drhodf_oxg[:, 4] = self.get_kinetic_energy_and_occd(ae)
        return rho_sxg, drhodf_oxg

    def get_paw_atom_contribs_occd(self, p_o, rho_sxg, drhodf_oxg):
        nspin = self.nspin
        norb = len(p_o)
        assert len(rho_sxg) == nspin
        x_gq = self.grids_indexer.empty_gq(nalpha=self.plan.nalpha)
        osize = self.plan.nalpha * (nspin + norb)
        x_orLq = self.grids_indexer.empty_rlmq(nalpha=osize)
        x_orLq.shape = (nspin + norb, x_orLq.shape[0], x_orLq.shape[1], x_gq.shape[-1])
        dat = [None] * nspin
        for s in range(nspin):
            rho_tuple = self.plan.get_rho_tuple(rho_sxg[s])
            arg_g, darg_g = self.plan.get_interpolation_arguments(rho_tuple, i=-1)
            fun_g, dfun_g = self.plan.get_function_to_convolve(rho_tuple)
            fun_g[:] *= self.w_g
            dfun_g = [dfun * self.w_g for dfun in dfun_g]
            p_gq, dp_gq = self.plan.get_interpolation_coefficients(arg_g.ravel(), i=-1)
            x_gq[:] = p_gq * fun_g[:, None]
            dat[s] = (darg_g, fun_g, dfun_g, p_gq, dp_gq)
            self.grids_indexer.reduce_angc_ylm_(x_orLq[s], x_gq, a2y=True, offset=0)
        for o in range(norb):
            s = p_o[o][0]
            x_gq[:] = 0
            darg_g, fun_g, dfun_g, p_gq, dp_gq = dat[s]
            drhodf_tuple = self.plan.get_drhodf_tuple(rho_sxg[s], drhodf_oxg[o])
            darg_df = 0
            dfun_df = 0
            for darg, dfun, drho in zip(darg_g, dfun_g, drhodf_tuple):
                darg_df += darg * drho
                dfun_df += dfun * drho
            darg_df[:] *= fun_g
            x_gq[:] = darg_df[..., None] * dp_gq + dfun_df[..., None] * p_gq
            self.grids_indexer.reduce_angc_ylm_(x_orLq[nspin + o], x_gq, a2y=True)
        return x_orLq

    def get_paw_atom_feat(self, rho_sxg, f_srLq):
        nspin, nx, ngrid = rho_sxg.shape
        nfeat = self.plan.nldf_settings.nfeat
        x_sig = np.zeros((nspin, nfeat, ngrid))
        f_gq = self.grids_indexer.empty_gq(nalpha=self.plan.nalpha)
        for s in range(nspin):
            self.grids_indexer.reduce_angc_ylm_(f_srLq[s], f_gq, a2y=False, offset=0)
            x_sig[s] = self.plan.eval_rho_full(
                f_gq,
                rho_sxg[s],
                # coeff_multipliers=self.plan.alpha_norms,
            )[0]
        return x_sig

    def get_paw_atom_feat_occd(self, p_o, rho_sxg, drhodf_oxg, f_srLq, f_orLq):
        nspin, nx, ngrid = rho_sxg.shape
        norb = len(p_o)
        nfeat = self.plan.nldf_settings.nfeat
        x_sig = np.zeros((nspin, nfeat, ngrid))
        dxdf_oig = np.zeros((norb, nfeat, ngrid))
        f_gq = self.grids_indexer.empty_gq(nalpha=self.plan.nalpha)
        dat = [None] * nspin
        for s in range(nspin):
            self.grids_indexer.reduce_angc_ylm_(f_srLq[s], f_gq, a2y=False, offset=0)
            x_sig[s] = self.plan.eval_rho_full(
                f_gq,
                rho_sxg[s],
                # coeff_multipliers=self.plan.alpha_norms,
            )[0]
            dat[s] = f_gq.copy()
        for o, p in enumerate(p_o):
            self.grids_indexer.reduce_angc_ylm_(f_orLq[o], f_gq, a2y=False, offset=0)
            s = p[0]
            dxdf_oig[o] = self.plan.eval_occd_full(
                dat[s],
                rho_sxg[s],
                f_gq,
                drhodf_oxg[o],
                # coeff_multipliers=self.plan.alpha_norms,
            )
        return x_sig, dxdf_oig


class PASDWCiderFeatureKernel(FastPASDWCiderKernel):
    PAWCiderContribs = PAWCiderFeatContribs

    def initialize_more_things(self, setups=None):
        self.nspin = self.dens.nt_sg.shape[0]
        if setups is None:
            setups = self.dens.setups
        for setup in setups:
            setup.cider_contribs = None
        for setup in setups:
            if not hasattr(setup, "cider_contribs") or setup.cider_contribs is None:
                setup.xc_correction = DiffPAWXCCorrection.from_setup(
                    setup, build_kinetic=self.is_mgga, ke_order_ng=False
                )
                # TODO it would be good to remove this
                setup.nlxc_correction = SBTGridContainer.from_setup(
                    # setup, rmin=setup.xc_correction.rgd.r_g[0]+1e-5, N=1024, encut=1e6, d=0.018
                    setup,
                    rmin=1e-4,
                    N=512,
                    encut=5e4,
                    d=0.02,
                )
                encut = setup.Z**2 * 20
                if encut - 1e-7 <= np.max(self.bas_exp_fit):
                    encut0 = np.max(self.bas_exp_fit)
                    Nalpha = self.bas_exp_fit.size
                else:
                    Nalpha = (
                        int(np.ceil(np.log(encut / self._amin) / np.log(self.lambd)))
                        + 1
                    )
                    encut0 = self._amin * self.lambd ** (Nalpha - 1)
                    assert encut0 >= encut - 1e-6, "Math went wrong {} {}".format(
                        encut0, encut
                    )
                    assert (
                        encut0 / self.lambd < encut
                    ), "Math went wrong {} {} {} {}".format(
                        encut0, encut, self.lambd, encut0 / self.lambd
                    )
                atom_plan = self.plan.new(nalpha=Nalpha)
                setup.cider_contribs = self.PAWCiderContribs.from_plan(
                    atom_plan,
                    self.plan,
                    self.cider_kernel,
                    setup.Z,
                    setup.xc_correction,
                    beta=1.6,
                    paw_algo=self._paw_algo,
                )
                if setup.cider_contribs.plan is not None:
                    assert (
                        abs(np.max(setup.cider_contribs.plan.alphas) - encut0) < 1e-10
                    )
                if self._paw_algo == "v1":
                    pss_cls = PSmoothSetupV1
                else:
                    pss_cls = PSmoothSetupV2
                setup.ps_setup = pss_cls.from_setup_and_atco(
                    setup,
                    setup.cider_contribs.atco_inp,
                    self.bas_exp_fit,
                    self.plan.alpha_norms,
                    encut0,
                    grid_nlm=setup.cider_contribs.nlm,
                )
                setup.pa_setup = PAugSetup.from_setup(setup)

    def calculate_paw_cider_features_p1(self, setups, D_asp, DD_aop, p_o):
        if len(D_asp.keys()) == 0:
            return {}, {}
        a0 = list(D_asp.keys())[0]
        nspin = D_asp[a0].shape[0]
        assert (nspin == 1) or self.is_cider_functional
        self.fr_asgLq = {}
        self.df_asgLq = {}
        self.c_asiq = {}

        for a, D_sp in D_asp.items():
            setup = setups[a]
            psetup = setup.ps_setup

            self.timer.start("coefs")
            rcalc = CiderRadialThetaDerivCalculator(setup.cider_contribs)
            expansion = CiderRadialDerivExpansion(rcalc, p_o)
            dx_sgLq, xt_sgLq = calculate_cider_paw_correction_deriv(
                expansion, setup, D_sp, DD_op=DD_aop[a]
            )
            dx_sgLq -= xt_sgLq
            self.timer.stop()
            self.timer.start("transform and convolve")
            fr_sgLq, df_sgLq, c_siq = setup.cider_contribs.calculate_y_terms(
                xt_sgLq, dx_sgLq, psetup
            )
            self.timer.stop()
            self.df_asgLq[a] = df_sgLq
            self.fr_asgLq[a] = fr_sgLq
            self.c_asiq[a] = c_siq
        return self.c_asiq, self.df_asgLq

    def calculate_paw_cider_features_p2(self, setups, D_asp, D_aoiq, DD_aop, p_o):
        if len(D_asp.keys()) == 0:
            if DD_aop is None or p_o is None:
                return {}, {}
            else:
                return {}, {}, {}, {}
        a0 = list(D_asp.keys())[0]
        nspin = D_asp[a0].shape[0]
        if DD_aop is None or p_o is None:
            norb = 0
        else:
            norb = DD_aop[a0].shape[0]
        self.vfr_asgLq = {}
        self.vdf_asgLq = {}
        ae_feat_asig = {}
        ps_feat_asig = {}
        ae_dfeat_aoig = {}
        ps_dfeat_aoig = {}

        for a, D_sp in D_asp.items():
            setup = setups[a]
            psetup = setup.pa_setup
            ni = psetup.ni
            Nalpha_sm = D_aoiq[a].shape[-1]
            dv_g = get_dv(setup.xc_correction.rgd)
            df_ogLq = self.df_asgLq[a]
            ft_ogLq = np.zeros_like(df_ogLq)
            fr_ogLq = self.fr_asgLq[a]
            # TODO C function for this loop? can go in pa_setup
            for i in range(ni):
                L = psetup.lmlist_i[i]
                n = psetup.nlist_i[i]
                Dref_sq = (psetup.pfuncs_ng[n] * dv_g).dot(fr_ogLq[:, :, L, :Nalpha_sm])
                j = psetup.jlist_i[i]
                for s in range(nspin + norb):
                    ft_ogLq[s, :, L, :Nalpha_sm] += (
                        D_aoiq[a][s, i] - Dref_sq[s]
                    ) * psetup.ffuncs_jg[j][:, None]
            ft_ogLq[:] += fr_ogLq
            if norb == 0:
                rcalc = CiderRadialFeatCalculator(setup.cider_contribs)
                expansion = CiderRadialExpansion(rcalc, ft_ogLq, df_ogLq)
                res = calculate_cider_paw_correction(
                    expansion, setup, D_sp, separate_ae_ps=True
                )
                ae_feat_asig[a], ps_feat_asig[a] = res
            else:
                rcalc = CiderRadialFeatDerivCalculator(setup.cider_contribs)
                expansion = CiderRadialDerivExpansion(rcalc, p_o, ft_ogLq, df_ogLq)
                res = calculate_cider_paw_correction_deriv(
                    expansion, setup, D_sp, DD_op=DD_aop[a]
                )
                ae_feat_asig[a], ae_dfeat_aoig[a] = res[0]
                ps_feat_asig[a], ps_dfeat_aoig[a] = res[1]
        if norb == 0:
            return ae_feat_asig, ps_feat_asig
        else:
            return ae_feat_asig, ps_feat_asig, ae_dfeat_aoig, ps_dfeat_aoig


def calculate_paw_sl_features(setups, D_asp, is_mgga):
    if len(D_asp.keys()) == 0:
        return {}, {}
    ae_feat_asig = {}
    ps_feat_asig = {}
    for a, D_sp in D_asp.items():
        setup = setups[a]
        rcalc = CiderRadialDensityCalculator(setup.xc_correction, D_sp, is_mgga)
        expansion = CiderRadialExpansion(rcalc)
        ae_feat_asig[a], ps_feat_asig[a] = calculate_cider_paw_correction(
            expansion, setup, D_sp, has_cider_contribs=False, separate_ae_ps=True
        )
    return ae_feat_asig, ps_feat_asig


def calculate_paw_sl_features_deriv(setups, D_asp, DD_aop, p_o, is_mgga):
    if len(D_asp.keys()) == 0:
        return {}, {}, {}, {}
    ae_feat_asig = {}
    ps_feat_asig = {}
    ae_dfeat_aoig = {}
    ps_dfeat_aoig = {}
    for a, D_sp in D_asp.items():
        setup = setups[a]
        rcalc = CiderRadialDensityDerivCalculator(
            setup.xc_correction, D_sp, DD_aop[a], is_mgga
        )
        expansion = CiderRadialDerivExpansion(rcalc, p_o)
        (
            (ae_feat_asig[a], ae_dfeat_aoig[a]),
            (ps_feat_asig[a], ps_dfeat_aoig[a]),
        ) = calculate_cider_paw_correction_deriv(
            expansion, setup, D_sp, DD_op=DD_aop[a], has_cider_contribs=False
        )
    return ae_feat_asig, ps_feat_asig, ae_dfeat_aoig, ps_dfeat_aoig


class CiderRadialThetaDerivCalculator:
    def __init__(self, xc):
        self.xc = xc
        self.mode = "step1"

    def __call__(
        self,
        rgd,
        p_o,
        n_sLg,
        dndf_oLg,
        Y_nL,
        dndr_sLg,
        dndrdf_oLg,
        rnablaY_nLv,
        ae,
    ):
        rho_sxg, drhodf_oxg = self.xc.vec_radial_vars_occd(
            n_sLg, dndf_oLg, Y_nL, dndr_sLg, dndrdf_oLg, rnablaY_nLv, ae
        )
        return self.xc.get_paw_atom_contribs_occd(p_o, rho_sxg, drhodf_oxg)


class CiderRadialFeatCalculator:
    def __init__(self, xc):
        self.xc = xc
        self.mode = "feature"

    def __call__(self, rgd, n_sLg, Y_nL, dndr_sLg, rnablaY_nLv, f_orLq, ae):
        rho_sxg = self.xc.vec_radial_vars(n_sLg, Y_nL, dndr_sLg, rnablaY_nLv, ae)[1]
        return self.xc.get_paw_atom_feat(rho_sxg, f_orLq)


class CiderRadialDensityCalculator:
    def __init__(self, xcc, D_sp, is_mgga):
        self.xcc = xcc
        self.D_sp = D_sp
        self.is_mgga = is_mgga
        self.mode = "feature"

    def __call__(self, rgd, n_sLg, Y_nL, dndr_sLg, rnablaY_nLv, ae):
        return vec_radial_vars_sl(
            self.xcc, n_sLg, Y_nL, dndr_sLg, rnablaY_nLv, ae, self.is_mgga, self.D_sp
        )[1]


class CiderRadialFeatDerivCalculator:
    def __init__(self, xc):
        self.xc = xc
        self.mode = "step2"

    def __call__(
        self,
        rgd,
        p_o,
        n_sLg,
        dndf_oLg,
        Y_nL,
        dndr_sLg,
        dndrdf_oLg,
        rnablaY_nLv,
        f_srLq,
        f_orLq,
        ae,
    ):
        rho_sxg, drhodf_oxg = self.xc.vec_radial_vars_occd(
            n_sLg, dndf_oLg, Y_nL, dndr_sLg, dndrdf_oLg, rnablaY_nLv, ae
        )
        return self.xc.get_paw_atom_feat_occd(p_o, rho_sxg, drhodf_oxg, f_srLq, f_orLq)


class CiderRadialDensityDerivCalculator:
    def __init__(self, xcc, D_sp, DD_op, is_mgga):
        self.xcc = xcc
        self.D_sp = D_sp
        self.DD_op = DD_op
        self.is_mgga = is_mgga
        self.mode = "rho"

    def __call__(self, n_sLg, dndf_oLg, Y_nL, dndr_sLg, dndrdf_oLg, rnablaY_nLv, ae):
        rho_sxg, drhodf_oxg = vec_radial_vars_occd(
            self.xcc,
            n_sLg,
            dndf_oLg,
            Y_nL,
            dndr_sLg,
            dndrdf_oLg,
            rnablaY_nLv,
            ae,
            self.D_sp,
            self.DD_op,
            self.is_mgga,
        )
        return rho_sxg, drhodf_oxg


class CiderRadialDerivExpansion:
    def __init__(self, rcalc, p_o, ft_orLq=None, df_orLq=None):
        self.rcalc = rcalc
        self.p_o = p_o
        self.ft_orLq = ft_orLq
        self.df_orLq = df_orLq
        assert self.rcalc.mode in ["step1", "step2", "rho"]

    def __call__(self, rgd, D_sLq, DD_oLq, n_qg, dndr_qg, nc0_sg, dnc0dr_sg, ae=True):
        n_sLg = np.dot(D_sLq, n_qg)
        n_sLg[:, 0] += nc0_sg
        dndf_oLg = np.dot(DD_oLq, n_qg)

        dndr_sLg = np.dot(D_sLq, dndr_qg)
        dndr_sLg[:, 0] += dnc0dr_sg
        dndrdf_oLg = np.dot(DD_oLq, dndr_qg)

        nspins, Lmax, nq = D_sLq.shape

        if self.rcalc.mode == "step1":
            wx_orLq = self.rcalc(
                rgd,
                self.p_o,
                n_sLg,
                dndf_oLg,
                Y_nL[:, :Lmax],
                dndr_sLg,
                dndrdf_oLg,
                rnablaY_nLv[:, :Lmax],
                ae,
            )
            return wx_orLq
        elif self.rcalc.mode == "step2":
            f_orLq = self.ft_orLq + self.df_orLq if ae else self.ft_orLq
            feat_xg, dfeat_xg = self.rcalc(
                rgd,
                self.p_o,
                n_sLg,
                dndf_oLg,
                Y_nL[:, :Lmax],
                dndr_sLg,
                dndrdf_oLg,
                rnablaY_nLv[:, :Lmax],
                f_orLq[:nspins],
                f_orLq[nspins:],
                ae,
            )
            return feat_xg, dfeat_xg
        elif self.rcalc.mode == "rho":
            feat_xg, dfeat_xg = self.rcalc(
                n_sLg,
                dndf_oLg,
                Y_nL[:, :Lmax],
                dndr_sLg,
                dndrdf_oLg,
                rnablaY_nLv[:, :Lmax],
                ae,
            )
            return feat_xg, dfeat_xg
        else:
            raise ValueError
